from qtpy import QtCore, QtGui, QtWidgets
from ..utils.ticks import tickValues, Ticks
      
fonts = []

# fonts.append(QtGui.QFont('Arial', 12))
# fonts.append(QtGui.QFont('Arial', 10))
# fonts.append(QtGui.QFont('Arial', 9))

fonts.append(QtGui.QFont('Arial', 8))
fonts.append(QtGui.QFont('Arial', 7))
fonts.append(QtGui.QFont('Arial', 5))

grid_pens = []
grid_pens.append(QtGui.QPen(QtGui.QColor(159,159,159), 0, QtCore.Qt.SolidLine))
grid_pens.append(QtGui.QPen(QtGui.QColor(191,191,191), 0, QtCore.Qt.DashLine))
grid_pens.append(QtGui.QPen(QtGui.QColor(223,223,223), 0, QtCore.Qt.DotLine))
        
        
class LabelItem(QtWidgets.QGraphicsLineItem):
    
    def __init__(self, text='', level=0, grid=False, parent=None, scene=None):
        super().__init__(parent=parent, scene=scene)
        self.setLine(0, 0, 0, 10)
        self.setFlags(QtWidgets.QGraphicsItem.ItemIgnoresTransformations)
        self.label = QtWidgets.QGraphicsTextItem(text, self)
        self.label.setFont(fonts[level])
        self.label.setPos(-1, 2)
        
        if grid:
            self.gline = QtWidgets.QGraphicsLineItem(self)
            self.gline.setPen(grid_pens[level])
            self.gline.setLine(0, -1e6, 0, 0)
            #self.gline.setZValue(0)
        
    def setRightAlign(self):
        self.label.setPos(2 - self.label.boundingRect().width(), 2)
                
                
class GridItem(QtWidgets.QGraphicsLineItem):
    
    def __init__(self, level=0,  parent=None, scene=None):
        super().__init__(parent=parent, scene=scene)
        self.setFlags(QtWidgets.QGraphicsItem.ItemIgnoresTransformations)
        self.gline = QtWidgets.QGraphicsLineItem(self)
        self.gline.setPen(grid_pens[level])
        self.gline.setLine(0, -1e6, 0, 0)    
        
        
class yAxisLabel(QtWidgets.QGraphicsLineItem):
    
    def __init__(self, text='', fontNumber=0, parent=None, scene=None):
        super().__init__(parent=parent, scene=scene)
        self.setFlags(QtWidgets.QGraphicsItem.ItemIgnoresTransformations)
        
        self.bgrect = QtWidgets.QGraphicsRectItem(-40, -10, 40, 20, parent=self)
        self.bgrect.setPen(QtGui.QPen(QtGui.QColor(250,250,250, 200)))
        self.bgrect.setBrush(QtGui.QBrush(QtGui.QColor(250,250,250, 200), QtCore.Qt.SolidPattern))        
        
        self.label = QtWidgets.QGraphicsTextItem(text, self)
        self.label.setDefaultTextColor(QtGui.QColor(120,120,120))
        self.label.setFont(fonts[fontNumber])
        self.label.setPos(-self.label.boundingRect().width(), -10)
        
    def setRightAlign(self):
        self.label.setPos(2 - self.label.boundingRect().width(), 2)        
        
        
class SubDivisionX(QtWidgets.QGraphicsLineItem):
    
    def __init__(self,  parent=None, scene=None):
        super().__init__(parent=parent, scene=scene)
        self.setLine(0, 0, 0, 3)
        self.setFlags(QtWidgets.QGraphicsItem.ItemIgnoresTransformations)


class SubDivisionY(QtWidgets.QGraphicsLineItem):
    
    def __init__(self,  parent=None, scene=None):
        super().__init__(parent=parent, scene=scene)
        self.setLine(0, 0, 3, 0)
        self.setFlags(QtWidgets.QGraphicsItem.ItemIgnoresTransformations)
               


#I think QGraphicsItemGroup is better, but i prevents moving the indicators
#I don't know why?
#class TickedRuler(QtWidgets.QGraphicsItemGroup):
class TickedRuler(QtWidgets.QGraphicsPolygonItem):
    
    def __init__(self, orientation, start, stop, scale, noDecimals=True, parent=None, scene=None):
        #super().__init__(parent=parent, scene=scene)   
        super().__init__(parent=parent)   
        
        #self.setPos(0,0)
        #self.setLine(-1, -1, 1, 1)
        #self.setPen(QtGui.QPen(QtGui.QColor(200,50,50),1))
        #self.setBrush(QtGui.QBrush(QtGui.QColor(200,50,50), QtCore.Qt.SolidPattern)) 
        
        self.orientation = orientation
        self.noDecimals = noDecimals        
        self.create_ticks(start, stop, scale)         
        self.init_bg()
        self.labelItems = dict()
        self.make_labels(self.ticks.push_values)
        
    def create_ticks(self, start, stop, scale):
        self.start = start
        self.stop = stop
        self.scale = scale    
        #self.thicks = tickValues(self.start, self.stop,  self.scale, 40, self.noDecimals)          
        self.ticks = Ticks(self.start, self.stop,  self.scale, 60, self.noDecimals)          
        
    @property
    def thicks(self):
        return self.ticks.values

    def update_labels(self, start, stop, scale, grid=False):
        self.start = start
        self.stop = stop
        self.scale = scale 
        
        self.ticks.update(start, stop, scale)                         
        self.remove_labels(self.ticks.pop_values)        
        self.make_labels(self.ticks.push_values, grid)
        
        if self.orientation == 0:
            self.axline.setLine(start, 0, stop, 0)
        else:
            self.axline.setLine(0, start, 0, stop)
        
    def init_bg(self):
        if self.orientation == 0:
            self.bgrect = QtWidgets.QGraphicsRectItem(-1e6, 0, 2e6, 22, parent=self)
            self.axline = QtWidgets.QGraphicsLineItem(self.start, 0, self.stop, 0, parent=self)
                
        if self.orientation == 90:
            self.bgrect = QtWidgets.QGraphicsRectItem(0, -1e6, 22, 2e6, parent=self)
            self.axline = QtWidgets.QGraphicsLineItem(0, self.start, 0, self.stop, parent=self)
            
        if self.orientation == -90:
            self.bgrect = QtWidgets.QGraphicsRectItem(-22, -1e6, 22, 2e6, parent=self)
            self.axline = QtWidgets.QGraphicsLineItem(0, self.start, 0, self.stop, parent=self)
            
        self.bgrect.setFlags(QtWidgets.QGraphicsItem.ItemIgnoresTransformations)
        self.bgrect.setPen(QtGui.QPen(QtGui.QColor(255,255,255), 0))
        self.bgrect.setBrush(QtGui.QBrush(QtGui.QColor(255,255,255), QtCore.Qt.SolidPattern))    
        
        self.axline.setPen(QtGui.QPen(QtGui.QColor(0,0,0), 0))
        

    def remove_labels(self, pop_values):
        if len(pop_values) == 0: 
            return
        for k in pop_values[0]:
            self.labelItems[k].setParentItem(None)
            self.labelItems.pop(k)        
        for k in pop_values[1]:
            self.labelItems[k].setParentItem(None)
            self.labelItems.pop(k)                    
        for k in pop_values[2]:
            self.labelItems[k].setParentItem(None)
            self.labelItems.pop(k)                    
                
    def make_labels(self, push_values, grid=False):
        if self.noDecimals:
            fmt = "%d"
        else:
            fmt = "%0.5g"                    
        
        if self.orientation == 0:
        
            for i in push_values[0]:
                i0 = LabelItem(fmt % i, 0, grid, self)
                i0.setPos(i, 0)
                self.labelItems[i] = i0
            
            for i in push_values[1]:
                i0 = LabelItem(fmt % i, 1, grid, self)
                i0.setPos(i, 0)
                self.labelItems[i] = i0
            
            for i in push_values[2]:        
                line = SubDivisionX(parent=self)
                line.setPos(i, 0)
                self.labelItems[i] = line
            
        if abs(self.orientation) == 90:
            for i in push_values[0]:
                i0 = LabelItem(fmt % i, 0, grid, self)            
                i0.setRightAlign()
                i0.setPos(0, i)
                i0.setRotation(-self.orientation)
                self.labelItems[i] = i0

            for i in push_values[1]:
                i0 = LabelItem(fmt % i, 1, grid, self)            
                i0.setRightAlign()
                i0.setPos(0, i)
                i0.setRotation(-self.orientation)
                self.labelItems[i] = i0
                
            for i in push_values[2]:        
                line = SubDivisionY(parent=self)
                line.setPos(0, i)
                self.labelItems[i] = line


#class Grid(QtWidgets.QGraphicsLineItem):
class Grid(QtWidgets.QGraphicsItemGroup):
    def __init__(self, ruler=None, parent=None, scene=None):
        #super().__init__(parent=parent, scene=scene)   
        super().__init__(parent=parent)   
        
        # self.setLine(-1, 1, 1, -1)
        # self.setPen(QtGui.QPen(QtGui.QColor(50,200,50),1))        
        
        self.ruler = ruler      
        self.labelItems = dict()
        self.make_labels(self.ticks.push_values)                
        
    @property
    def ticks(self):
        return self.ruler.ticks

    @property        
    def orientation(self):
        return self.ruler.orientation

    def update_labels(self, grid=True):                        
        self.remove_labels(self.ticks.pop_values)        
        self.make_labels(self.ticks.push_values, grid)
          

    def remove_labels(self, pop_values):
        if len(pop_values) == 0: 
            return
        for k in pop_values[0]:
            self.labelItems[k].setParentItem(None)
            self.labelItems.pop(k)        
        for k in pop_values[1]:
            self.labelItems[k].setParentItem(None)
            self.labelItems.pop(k)                    
        for k in pop_values[2]:
            self.labelItems[k].setParentItem(None)
            self.labelItems.pop(k)                    
                
    def make_labels(self, push_values, grid=False):                         
        if self.orientation == 0:        
            for i in push_values[0]:
                i0 = GridItem(0, self)
                i0.setPos(i, 0)
                self.labelItems[i] = i0
            
            for i in push_values[1]:
                i0 = GridItem(1, self)
                i0.setPos(i, 0)
                self.labelItems[i] = i0
            
            for i in push_values[2]:        
                line = GridItem(2, self)
                line.setPos(i, 0)
                self.labelItems[i] = line
            
        if abs(self.orientation) == 90:
            for i in push_values[0]:
                i0 = GridItem(0, self)            
                i0.setPos(0, i)
                i0.setRotation(-self.orientation)
                self.labelItems[i] = i0

            for i in push_values[1]:
                i0 = GridItem(1, self)           
                i0.setPos(0, i)
                i0.setRotation(-self.orientation)
                self.labelItems[i] = i0
                
            for i in push_values[2]:        
                line = GridItem(2, self)
                line.setPos(0, i)
                line.setRotation(-self.orientation)
                self.labelItems[i] = line                  
               
            
class Axis(QtWidgets.QGraphicsLineItem):
    def __init__(self, plotAngle, start, stop, thicks, parent=None, scene=None):
        super().__init__(parent=parent, scene=scene)   
        
        self.setLine(0, 0, 0, 0)
        
        self.start = start
        self.stop = stop
        self.thicks = thicks
        self.plotAngle = plotAngle        
        
        self.createAxis()
            
    def createAxis(self):   
        if self.plotAngle == 0:
            
            for thickLevel in range(len(self.thicks)):
                if self.thicks[thickLevel][0] > 15:
                    for i in self.thicks[thickLevel][1]:
                        label = yAxisLabel('%0.5g' % i, thickLevel, self)
                        label.setPos(0, i)
            
        else:
            
            for thickLevel in range(len(self.thicks)):
                if self.thicks[thickLevel][0] > 15:
                    for i in self.thicks[thickLevel][1]:
                        label = yAxisLabel('%0.5g' % i, thickLevel, self)
                        label.setPos(i, 0)
                        label.setRotation(-90)



import sys
import math
import logging

import numpy as np

try:
    from ...utils import numba_func
    has_numba = True
except:
    has_numba = False    
    
logger = logging.getLogger(__name__)

def is_integer_num(n):
    if isinstance(n, int):
        return True
    if isinstance(n, float):
        return n.is_integer()
    return False
    
def natural_range(dtype):
    if dtype in ['uint8', 'int8']:
        return 256
    elif dtype in ['uint16', 'int16']:
        return 65536
    elif dtype in ['uint32', 'int32']:
        return 4294967296
    elif dtype in ['uint64', 'int64']:
        return 18446744073709551616
    elif dtype in ['float16', 'float32', 'float64']:
        return 1
    else:
        raise TypeError('dtype %s not supported to display' % str(statbuff.dtype))
        
def get_map16(dtype='float64', offset=0, gain=1):
    natrange = natural_range(dtype)
    return ((np.arange(natrange) - offset) * gain * 65536 / natrange).clip(0, 65536).astype('uint16')    

def hist2d(array, bins=64, step=None, low=None, high=None, pow2snap=True, plot=False, use_numba=True):
    if array.dtype in ['int8', 'uint8', 'int16', 'uint16']:
        assert pow2snap
        hist, starts, stepsize = hist16bit(array, bins, step, low, high, use_numba=True)
        return hist, starts, stepsize
        
    elif array.dtype in ['float16', 'float32', 'float64']:
        hist, starts, stepsize = histfloat(array, bins, step, low, high, pow2snap)
        return hist, starts, stepsize        
        
    if plot:
        plt.bar(starts + stepsize/2, hist, stepsize)
        plt.grid()
        plt.xlabel('value [DN]')
        plt.ylabel('count')        
    
    return hist, starts, stepsize
    
def hist16bit(array, bins=64, step=None, low=None, high=None, use_numba=True):
    """
    stepsize should be power of 2    
    array should be 8 or 16 bit integer
    """               
    if array.dtype in ['int8','uint8']:
        length = 256
    elif array.dtype in ['int16', 'uint16']:
        length = 65536
    
    if use_numba and has_numba:
        hist = numba_func.bincount2d(array, length)    
    else:
        if use_numba:
            logger.warning('Numba is not available')
        hist = np.bincount(array.ravel(), minlength=length)    
    
    non_zeros_indices = np.argwhere(hist > 0)
    min_index = non_zeros_indices[0][0]
    max_index = non_zeros_indices[-1][0]
    
    first_edge = min_index if low is None else low  
    last_edge = max_index if high is None else high
    
    if first_edge == last_edge:
        first_edge -= 0.5
        last_edge += 0.5

    if step is None:    
        stepsize = max((last_edge - first_edge) / (bins-1), 1)
    else:
        stepsize = step               
    
    stepsize = max(2**math.floor(np.log2(stepsize)), 1)    
    
    if stepsize > 1:
        hist = hist.reshape(length // stepsize, stepsize).sum(1)
        non_zeros_indices = np.argwhere(hist > 0)
        min_index = non_zeros_indices[0][0]
        max_index = non_zeros_indices[-1][0]
    
    starts = np.arange(min_index, max_index+1) * stepsize   
    
    return hist[min_index:max_index+1], starts, stepsize
    
def histfloat(array, bins=64, step=None, low=None, high=None, pow2snap=True, use_numba=True):

    if (low is None or high is None):
        if use_numba:
            minimum, maximum = numba_func.get_min_max(array)
        else:
            minimum, maximum = array.min(), array.max()
        
    first_edge = minimum if low is None else low    
    last_edge = maximum if high is None else high    

    if first_edge == last_edge:
        first_edge -= 0.5
        last_edge += 0.5
    
    if step is None:    
        stepsize = (last_edge - first_edge) / (bins-1)
    else:
        stepsize = step               
    
    if pow2snap:
        stepsize = 2**math.floor(np.log2(stepsize))
    
    start_ind = math.floor(first_edge / stepsize) 
    end_ind = math.floor(last_edge / stepsize) + 1
    
    starts = np.arange(start_ind, end_ind) * stepsize
    offset = starts[0]
      
    assert len(starts) <= 65536
    
    #TO DO, clipping is only needed if values are outside the bins
    #This means that minimum and maximum should always be calculated
    if (offset != 0) and (stepsize != 1):
        array16bit = ((array - offset) / stepsize).clip(0, 65536).astype('uint16') 
    elif (stepsize != 1):    
        array16bit = (array / stepsize).clip(0, 65536).astype('uint16') 
    elif (offset != 0):        
        array16bit = (array - offset).clip(0, 65536).astype('uint16') 
    
    if use_numba and has_numba:
        hist = numba_func.bincount2d(array16bit, 65536)[:len(starts)] 
    else:
        hist = np.bincount2d(array16bit, minlength=65536)[:len(starts)]        
    
    return hist, starts, stepsize  
    



import logging
from pathlib import Path
import json
from collections import OrderedDict

from qtpy import QtCore, QtGui, QtWidgets
from qtpy.QtCore import Qt

from jinja2 import Template

from ... import gui, config
from ...panels.base import BasePanel, CheckMenu
from ...dialogs.formlayout import fedit

logger = logging.getLogger(__name__)
respath = Path(config['respath'])

class FileLayout(QtWidgets.QWidget):
    """File-specialized QLineEdit layout"""
    def __init__(self, text='', loadCall=None, saveCall=None, save=False, edit=True):
        super().__init__()
        
        self.loadCall = loadCall         
        self.saveCall = saveCall
        self.save = save
        
        hbox = QtWidgets.QHBoxLayout()
        hbox.setMargin(0)
        hbox.setContentsMargins(0,0,0,0)
        self.setLayout(hbox)        
                
        self.fileExplBtn = QtWidgets.QToolButton()
        self.fileExplBtn.clicked.connect(self.fileExplore)
        self.fileExplBtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'folders_explorer.png')))
        hbox.addWidget(self.fileExplBtn)
        
        self.lineedit = QtWidgets.QLineEdit(text)
        hbox.addWidget(self.lineedit)
        
        if self.loadCall is None:
            self.loadCall = lambda filepath: None 
            
        else:
            self.fileLoadbtn = QtWidgets.QToolButton()        
            self.fileLoadbtn.clicked.connect(self.loadFile)
            self.fileLoadbtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'folder.png')))
            hbox.addWidget(self.fileLoadbtn)
        
        if self.saveCall is None:
            self.saveCall = lambda filepath: None
            
        else:
            self.fileSavebtn = QtWidgets.QToolButton()        
            self.fileSavebtn.clicked.connect(self.savefile)
            self.fileSavebtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'diskette.png')))
            hbox.addWidget(self.fileSavebtn)
        
        if edit:
            self.editbtn = QtWidgets.QToolButton()
            self.editbtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'script_edit.png')))
            self.editbtn.clicked.connect(self.editfile)
            hbox.addWidget(self.editbtn)      

    def fileExplore(self):            
        selectedFile = gui.dialog.selectfile(default_path=self.text())
        
        if selectedFile is None: return        
            
        self.lineedit.setText(selectedFile)                        
            
        if not self.save:
            self.loadCall(Path(self.text()))
        
    def loadFile(self):
        self.loadCall(Path(self.lineedit.text()))

    def savefile(self):                  
        self.saveCall(Path(self.lineedit.text()))
            
    def editfile(self):
        from gdesk import shell
        shell.edit_file(self.lineedit.text())

    def text(self):
        return self.lineedit.text()  
        
    def setText(self, text):
        self.lineedit.setText(text)
        
class DirFileLayout(QtWidgets.QWidget):
    """File-specialized QLineEdit layout"""
    def __init__(self, fullFilePath='', loadCall=None, saveCall=None, save=False, edit=True):
        super().__init__()
        
        self.loadCall = loadCall         
        self.saveCall = saveCall
        
        hbox = QtWidgets.QHBoxLayout()
        hbox.setMargin(0)
        hbox.setContentsMargins(0,0,0,0)
        self.setLayout(hbox)        
                
        self.dirExplBtn = QtWidgets.QToolButton()
        self.dirExplBtn.clicked.connect(self.getMap)
        self.dirExplBtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'folders_explorer.png')))
        hbox.addWidget(self.dirExplBtn)
        
        fullFilePath = Path(fullFilePath)
        
        self.dirEdit = QtWidgets.QLineEdit(str(fullFilePath.parent))
        hbox.addWidget(self.dirEdit, 2)
        
        self.fileNameEdit = QtWidgets.QLineEdit(fullFilePath.name)
        hbox.addWidget(self.fileNameEdit, 1)        
        
        if self.loadCall is None:
            self.loadCall = lambda filepath: None 
            
        else:
            self.fileLoadbtn = QtWidgets.QToolButton()        
            self.fileLoadbtn.clicked.connect(self.loadFile)
            self.fileLoadbtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'folder_vertical_document.png')))
            hbox.addWidget(self.fileLoadbtn)
        
        if self.saveCall is None:
            self.saveCall = lambda filepath: None
            
        else:
            self.fileSavebtn = QtWidgets.QToolButton()        
            self.fileSavebtn.clicked.connect(self.savefile)
            self.fileSavebtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'diskette.png')))
            hbox.addWidget(self.fileSavebtn)
        
        if edit:
            self.editbtn = QtWidgets.QToolButton()
            self.editbtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'script_edit.png')))
            self.editbtn.clicked.connect(self.editfile)
            hbox.addWidget(self.editbtn)

    @property
    def fullFileName(self):
        return Path(self.dirEdit.text()) / self.fileNameEdit.text()            

    def getMap(self):
        filepath = Path(gui.dialog.selectfile())     
        self.dirEdit.setText(str(filepath.parent))
        self.fileNameEdit.setText(filepath.name)
        
    def loadFile(self):
        fileName = self.fileNameEdit.text()
        
        if fileName.strip() == '':
            fullFileName = Path(gui.getfilename(file=self.dirEdit.text()))
            self.dirEdit.setText(str(fullFileName.parent))
            self.fileNameEdit.setText(str(fullFileName.name))
        
        self.loadCall(self.fullFileName)

    def savefile(self):
        fileName = self.fileNameEdit.text()
        
        if fileName.strip() == '':
            fullFileName = Path(gui.putfilename())
            self.dirEdit.setText(str(fullFileName.parent))
            self.fileNameEdit.setText(str(fullFileName.name))
            
        elif self.fullFileName.exists():
            if not gui.question('File exists. Do you want to overwrite?'):                
                fullFileName = Path(gui.putfilename())
                self.dirEdit.setText(str(fullFileName.parent))
                self.fileNameEdit.setText(str(fullFileName.name))                
            
        self.saveCall(self.fullFileName)
            
    def editfile(self):
        from gdesk import shell
        shell.edit_file(self.fullFileName)

    def text(self):
        return str(self.fullFileName)
        
    def setText(self, text):
        fullFileName = Path(text)
        self.dirEdit.setText(str(fullFileName.parent))
        self.fileNameEdit.setText(str(fullFileName.name))      

class MyTable(QtWidgets.QTableWidget):

    def __init__(self, headers):
        self.headers = headers
        colcount = len(headers)
        super().__init__(1, colcount)  
        self.setHorizontalHeaderLabels(self.headers)
        
class FormFile(FileLayout):

    def __init__(self, cfg):
        super().__init__(text=cfg.get('text', ''), edit=cfg.get('edit', False))

    def setContent(self, content):
        self.setText(content)
        
    def getContent(self):
        return self.text()
        
        
class FormLineEdit(QtWidgets.QLineEdit):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    def setContent(self, content):
        self.setText(content)
        
    def getContent(self):
        return self.text()    
        

class FormTextEdit(QtWidgets.QPlainTextEdit):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWordWrapMode(QtGui.QTextOption.NoWrap)
        
    def setContent(self, content):
        self.setPlainText(content)
        
    def getContent(self):
        return self.toPlainText()            
        

class FormTable(QtWidgets.QWidget):

    def __init__(self, cfg):
        super().__init__()
        vbox = QtWidgets.QVBoxLayout()
        vbox.setContentsMargins(0,0,0,0)
        vbox.setMargin(0)
        self.setLayout(vbox)
        
        hbox = QtWidgets.QHBoxLayout()
        vbox.addLayout(hbox)                      
        
        self.addBtn = QtWidgets.QToolButton()
        self.addBtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'table_row_insert.png')))
        self.addBtn.clicked.connect(self.addItem)
        hbox.addWidget(self.addBtn)
        
        self.deleteBtn = QtWidgets.QToolButton()
        self.deleteBtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'table_row_delete.png')))
        self.deleteBtn.clicked.connect(self.deleteItem)
        hbox.addWidget(self.deleteBtn)
        
        self.moveUpBtn = QtWidgets.QToolButton()
        self.moveUpBtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'table_rows_insert_above_word.png')))
        self.moveUpBtn.clicked.connect(self.moveUp)
        hbox.addWidget(self.moveUpBtn)
        
        self.moveDownBtn = QtWidgets.QToolButton()
        self.moveDownBtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'table_rows_insert_below_word.png')))
        self.moveDownBtn.clicked.connect(self.moveDown)
        hbox.addWidget(self.moveDownBtn)        
        
        # self.lineEdit = QtWidgets.QLineEdit()
        # hbox.addWidget(self.lineEdit)
        hbox.addStretch()
                
        self.table = MyTable(cfg.get('headers', ['None']))
        vbox.addWidget(self.table)
        
    def addItem(self):
        self.table.setRowCount(self.table.rowCount() + 1)
        
    def deleteItem(self):
        self.table.setRowCount(self.table.rowCount() - 1) 

    def moveUp(self):
        rownr = self.table.selectedIndexes()[0].row()
        if rownr > 0:
            self.swapRow(rownr, rownr-1) 
            self.table.selectRow(rownr-1)
        
    def moveDown(self):
        rownr = self.table.selectedIndexes()[0].row()
        if rownr < (self.table.rowCount() - 1):
            self.swapRow(rownr, rownr+1)
            self.table.selectRow(rownr+1)
        
    def swapRow(self, rowNr0, rowNr1):
        row0 = self.takeRow(rowNr0)
        row1 = self.takeRow(rowNr1)
        self.setRow(rowNr0, row1)
        self.setRow(rowNr1, row0)
        
    def takeRow(self, rownr):
        row = []
        for colnr in range(self.table.columnCount()):
            row.append(self.table.takeItem(rownr, colnr))
        return row
        
    def setRow(self, rownr, row):
        for colnr, item in enumerate(row):
            self.table.setItem(rownr, colnr, item)
        return row        
        
    def getContent(self):
        content = []
        table = self.table
        for r in range(table.rowCount()):
            item = dict()
            for c in range(table.columnCount()):
                header = table.horizontalHeaderItem(c).text()
                cell = table.item(r, c)
                if not cell is None:
                    text = cell.text()
                    item[header] = text
                else:
                    item[header] = None
            content.append(item)
        return content
        
    def setContent(self, content):
        table = self.table
        table.setRowCount(len(content))
        
        for r, item in enumerate(content):
            for key, val in item.items():
                c = table.headers.index(key)
                table.setItem(r, c, QtWidgets.QTableWidgetItem(val))
      
        table.resizeColumnsToContents()


class CustomForm(QtWidgets.QWidget):
    def __init__(self, cfg):
        super().__init__()
        
        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)       
        
        self.flay = flay = QtWidgets.QFormLayout()
        flay.setVerticalSpacing(5)
        self.setLayout(flay)

        self.widgets = OrderedDict()
        self.widgetsMeta = dict()                
        
        self.setWidgets(cfg)        
        
    def setWidgets(self, cfg):        
        for item in cfg:
            self.addFormItem(item['key'], item['caption'], '', item['widget_type'],
                item.get('description', None),
                item.get('widget_config', {}))
        
    def getWidgetsConfig(self):
        widgetsConfig = list()
        
        for key, widget in self.widgets.items():
            meta = self.widgetsMeta[key]
            item = {'key': key, 'caption': meta['caption'], 'widget_type': meta['widget_type']}
            widgetsConfig.append(item)
            
        return widgetsConfig
        
    def addFormItem(self, key, caption, content=None, widgetType='FormLineEdit', description=None, widget_cfg=None):
        if widgetType == 'FormLineEdit':
            widget = FormLineEdit()
            widget.setContent(content)
            
        elif widgetType == 'FormTextEdit':
            widget = FormTextEdit()
            widget.setContent(content)
            
        elif widgetType == 'FormTable':
            widget = FormTable(widget_cfg)  

        elif widgetType == 'FormFile':
            widget = FormFile(widget_cfg)     
            
        self.widgets[key] = widget
        self.widgetsMeta[key] = {'caption': caption, 'widget_type': widgetType}
        captionLabel = QtWidgets.QLabel(caption)
        if not description is None:
            captionLabel.setToolTip(description)
        self.flay.addRow(captionLabel, widget)        
                
    def setContent(self, cfg):
        for key, widget in self.widgets.items():
            if not key in cfg.keys():
                logger.warn(f'{key} not found')
                continue
            widget.setContent(cfg[key])
        
    def getContent(self):                
        cfg = dict()        
        
        for key, widget in self.widgets.items():
            cfg[key] = widget.getContent()                                
        
        return cfg  

class TemplateScrollForm(QtWidgets.QScrollArea):        
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWidgetResizable(True)        
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)            
        
    def setWidgets(self, cfg):
        self.form = CustomForm(cfg)
        self.setWidget(self.form)        
        
    # def setContent(self, cfg):
        # self.form.setContent(cfg)
        
    # def getContent(self):
        # return self.form.getContent()                    

        
class TemplateWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        vbox = QtWidgets.QVBoxLayout()
        vbox.setContentsMargins(5,5,5,5)
        vbox.setMargin(5)
        self.setLayout(vbox)

        flay = QtWidgets.QFormLayout()
        vbox.addLayout(flay)        

        self.formConfig = QtWidgets.QLineEdit()
        self.formConfig.setReadOnly(True)
        flay.addRow('Form', self.formConfig)        

        self.settingsFileName = DirFileLayout('',
            loadCall=self.loadValues, saveCall=self.saveValues)
        flay.addRow('Settings', self.settingsFileName)

        self.templateForm = TemplateScrollForm(self)
        vbox.addWidget(self.templateForm)

        flay = QtWidgets.QFormLayout()
        vbox.addLayout(flay)

        self.codeTemplate = QtWidgets.QLineEdit()
        self.codeTemplate.setReadOnly(True)
        flay.addRow('Script Template', self.codeTemplate)

        self.scriptFileName = FileLayout('', save=True)
        flay.addRow('Script', self.scriptFileName)      

        hbox = QtWidgets.QHBoxLayout()
        hbox.setMargin(0)
        vbox.addLayout(hbox)

        self.makeBtn = QtWidgets.QPushButton('Make')
        self.makeBtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'script_lightning.png')))
        self.makeBtn.clicked.connect(self.make)            

        hbox.addWidget(self.makeBtn)
        
    @property
    def form(self):
        return self.templateForm.widget()
        
    def loadValues(self, filepath):
        if filepath.suffix.lower() in ['.toml']:
        
            import toml        
            cfg = toml.load(str(filepath))
            
        elif filepath.suffix.lower() in ['.json']: 
        
            with open(filepath, 'r') as fp:
                cfg = json.load(fp)
            
        self.form.setContent(cfg['params'])
        
        for call in cfg['calls']:
            self.runTab.setFormWidgets(call['call'], updateCombo=True)        
            self.runTab.form.setContent(call['kwargs'])        
            self.runTab.returnRef.setText(call['retref'])        
        
    def saveValues(self, filepath):
        cfg = dict()
        cfg['params'] = self.form.getContent()
        cfg['calls'] = [{'call': self.runTab.funcName.currentText(), 'retref': self.runTab.returnRef.text(), 'kwargs': self.runTab.form.getContent()}]
        
        if filepath.suffix.lower() in ['.toml']:
        
            import toml           
            with open(filepath, 'w') as fp:
                toml.dump(cfg, fp)
                
        elif filepath.suffix.lower() in ['.json']:  
            
            with open(filepath, 'w') as fp:
                json.dump(cfg, fp,  indent='    ')
            
    @property    
    def runTab(self):
        return self.parent().parent().runTab
                
    def make(self):                
        templateFileName = self.codeTemplate.text()                
        
        cfg = self.form.getContent()        
        scriptFileName = self.scriptFileName.text()                               
        
        templ_grab = Template(open(templateFileName, 'r').read())                        
        pycode = templ_grab.render(**cfg)
    
        with open(scriptFileName, 'w') as fp:
            fp.write(f'# Created by ninja2 from template {templateFileName}\n')
            fp.write(pycode)
            
        self.runTab.scriptFileWidget.setText(scriptFileName)
            
            
class RunScrollForm(QtWidgets.QScrollArea):        
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWidgetResizable(True)        
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)            
        
    def setWidgets(self, cfg):
        self.form = CustomForm(cfg)
        self.setWidget(self.form)        
        
    def setContent(self, cfg):
        self.form.setContent(cfg)
        
    def getContent(self):
        return self.form.getContent()              
            
        
class RunWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)        
        
        vbox = QtWidgets.QVBoxLayout()
        vbox.setContentsMargins(5, 5, 5, 5)
        vbox.setMargin(5)
        self.setLayout(vbox)
        
        flay = QtWidgets.QFormLayout()
        vbox.addLayout(flay)         
        
        self.scriptFileWidget = FileLayout('')
        flay.addRow('Script', self.scriptFileWidget)        
        
        hbox = QtWidgets.QHBoxLayout()
        self.funcName = QtWidgets.QComboBox()
        self.funcName.setEditable(False)
        self.funcName.currentIndexChanged.connect(self.funcNameIndexChanged)
        hbox.addWidget(self.funcName)
        self.locateBtn = QtWidgets.QPushButton('Locate')
        self.locateBtn.clicked.connect(self.locate)
        hbox.addWidget(self.locateBtn)                
        
        flay.addRow('Function', hbox)        
        
        self.callerEdit = QtWidgets.QLineEdit('')
        flay.addRow('Caller', self.callerEdit)
        
        self.settingsFileName = FileLayout('',
            loadCall=self.loadValues, saveCall=self.saveValues)
        flay.addRow('Settings', self.settingsFileName)        
        
        self.runScrollForm = RunScrollForm()
        vbox.addWidget(self.runScrollForm)
        
        flay = QtWidgets.QFormLayout()
        vbox.addLayout(flay)         
        
        self.returnRef = QtWidgets.QLineEdit()
        flay.addRow('Return Reference', self.returnRef)        
        
        hbox = QtWidgets.QHBoxLayout()
        vbox.addLayout(hbox)                                    
        
        self.runBtn = QtWidgets.QPushButton('Run')
        self.runBtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'script_go.png')))
        self.runBtn.clicked.connect(self.run)
        hbox.addWidget(self.runBtn)   

        self.widgetConfigs = dict()

    @property
    def form(self):
        return self.runScrollForm.widget()

    def addFormWidgetsConfig(self, funcName, cfg):
        self.widgetConfigs[funcName] = cfg
        self.funcName.addItem(funcName)
        
    def setFormWidgets(self, funcName, updateCombo=False):
        cfg = self.widgetConfigs[funcName]
        self.runScrollForm.setWidgets(cfg)  
        if updateCombo:
            index = self.funcName.findText(funcName)        
            self.funcName.setCurrentIndex(index)
        
    def clearFormWidgets(self):        
        while self.funcName.count() > 0:
            self.funcName.removeItem(0)
        self.widgetConfigs.clear()

    @property
    def scriptFileName(self):        
        return self.scriptFileWidget.text()
        
    def funcNameIndexChanged(self, *args):
        funcName = self.funcName.currentText()
        if funcName == '': return
        self.setFormWidgets(funcName)
        
    def locate(self):
        from gdesk import shell, use
        
        scriptFileName = Path(self.scriptFileName).resolve()
        funcName = self.funcName.currentText()
        console = gui.qapp.panels.selected('console')
        
        for path in use.__script_manager__.path:
            try:
                use_path = scriptFileName.relative_to(Path(path).resolve())
                break
            except:
                pass    
        else:
            logger.warn('Could not locate script')

        self.callerEdit.setText(f'{use_path.stem}.{funcName}')


    def loadValues(self, filepath):
        if filepath.suffix.lower() in ['.toml']:
        
            import toml        
            cfg = toml.load(str(filepath))
            
        elif filepath.suffix.lower() in ['.json']: 
        
            with open(filepath, 'r') as fp:
                cfg = json.load(fp)
            
        self.form.setContent(cfg)
        
    def saveValues(self, filepath):
        cfg = self.form.getContent()
        
        if filepath.suffix.lower() in ['.toml']:
            import toml           
            with open(filepath, 'w') as fp:
                toml.dump(cfg, fp)
                
        elif filepath.suffix.lower() in ['.json']:            
            with open(filepath, 'w') as fp:
                json.dump(cfg, fp,  indent='    ')
        
        
    def run(self):
        from gdesk import shell, use
        
        cfg = self.form.getContent()
        
        console = gui.qapp.panels.selected('console')        
        self.locate()                 
            
        module_func = self.callerEdit.text()
        parts = module_func.split('.')
        module, func = '.'.join(parts[:-1]), parts[-1]            

        retref = self.returnRef.text().strip()
        
        console.exec_cmd('use.__script_manager__.update_now()')

        if not retref == '':
            console.exec_cmd(f'{retref} = use("{module}").{func}(**{cfg})')
        else:
            console.exec_cmd(f'use("{module}").{func}(**{cfg})')
        
        
class ScriptTabs(QtWidgets.QTabWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)      

        self.templateTab = TemplateWidget()
        self.runTab = RunWidget()
        
        self.addTab(self.templateTab, 'Template')
        self.addTab(self.runTab, 'Run')
        
    @property
    def scriptFileName(self):
        return self.templateTab.scriptFileName.text()
        
class WrapCaller(object):
    def __init__(self, caller, *args, **kwargs):
        self.caller = caller
        self.args = args
        self.kwargs = kwargs
                
    def __call__(self):
        self.caller(*self.args, **self.kwargs)
        
class RecentMenu(QtWidgets.QMenu):
    def __init__(self, parent=None):
        super().__init__('Recent', parent)
        self.scriptPanel = self.parent()
        #self.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'images.png')))
        
    def showEvent(self, event):
        self.initactions()        

    def initactions(self):
        self.clear()        
        self.actions = []

        for rowid, timestamp, path in gui.qapp.history.yield_recent_paths(category='scriptwiz'):                                
            action = QtWidgets.QAction(path, self)
            action.triggered.connect(WrapCaller(self.scriptPanel.openTemplate, path))
            self.addAction(action)
            self.actions.append(action)         

class ScriptWizardPanel(BasePanel):
    panelCategory = 'scriptwiz'
    panelShortName = 'basic'
    userVisible = True
    
    classIconFile = str(respath / 'icons' / 'px16' / 'script_bricks.png')

    def __init__(self, parent, panid):
        super().__init__(parent, panid, type(self).panelCategory)
        
        self.fileMenu = CheckMenu("&File", self.menuBar())
        
        self.addMenuItem(self.fileMenu, 'Open Template', self.openTemplateDialog, 
            statusTip="Open a script template")
            
        self.fileMenu.addMenu(RecentMenu(self))  
            
        self.addMenuItem(self.fileMenu, 'Load Values', self.loadValues, 
            statusTip="Load Values")     

        self.addMenuItem(self.fileMenu, 'Save Values', self.saveValues, 
            statusTip="Save Values")             
        
        self.addMenuItem(self.fileMenu, 'Close', self.close_panel, 
            statusTip="Close this levels panel",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'cross.png')))        
        
        self.tabs = ScriptTabs()
        self.setCentralWidget(self.tabs)
        
        self.addBaseMenu()
        self.statusBar().hide()
        
    @property
    def templateTab(self):
        return self.tabs.templateTab
        
    @property
    def runTab(self):
        return self.tabs.runTab
        
    def openTemplateDialog(self):
        filename = gui.getfilename('*.json', 'Open Script Template')
        self.openTemplate(filename)
    
    def openTemplate(self, filename):    
        from gdesk import shell, use
        
        scm = use.__script_manager__
        filename = Path(filename)              
        
        cfg = json.load(open(filename, 'r'))
        
        gui.qapp.history.storepath(str(filename), category='scriptwiz')
        
        self.templateTab.formConfig.setText(str(filename))        
        self.templateTab.templateForm.setWidgets(cfg.get('form', {}))           
        
        self.templateTab.settingsFileName.setText(str(filename.parent / cfg.get('settings_file', '')))
        self.templateTab.codeTemplate.setText(str(filename.parent / cfg['code_template']))        
        self.templateTab.scriptFileName.setText(str(Path(scm.path[0]) / cfg.get('script', '')))
        
        self.runTab.clearFormWidgets()
        
        calls = cfg.get('calls', [])        
        
        for call in calls:       
            funcName = call.get('call')            
            self.runTab.addFormWidgetsConfig(funcName, call.get('form', {}))
            
        default_call = cfg.get('default_call', None)
        index = self.runTab.funcName.findText(default_call)
        self.runTab.funcName.setCurrentIndex(index)
        
    def loadValues(self):
        import toml
        valuesFile = gui.getfilename('*.toml')
        cfg = toml.load(valuesFile)
        self.templateTab.form.setContent(cfg)
        
    def saveValues(self):
        import toml
        valuesFile = gui.putfilename('*.toml')
        cfg = self.templateTab.form.getContent()
        with open(valuesFile, 'w') as fp:
            toml.dump(cfg, fp)



import os
import sys
import threading
import pathlib
import pprint

from multiprocessing import Process, Queue

from .. import config, configure, gui
configure(matplotlib={'backend':'svg'})

from .. import console
from ..core.shellmod import Shell
from ..core.tasks import ZmqQueues
from ..core.tasks import CommQueues
from ..core.watcher import CommandClient


def python_executable():
    executable = pathlib.Path(sys.executable)
    if executable.name == 'python.exe':
        #This is also correct for virtual env
        #Note that the os module is loaded from the master Python location
        #not the virtual env
        return str(executable)
    else:
        #In case of embeded python in on other executable (Canvas)
        #Base the python.exe location on the os module location
        executable = pathlib.Path(os.__file__).parent.parent / 'python.exe'
        return str(executable)


def connect_to_gui(port=None, host='localhost',
        namespace=None, gui_redirect=True, as_server=False):
    """
    Start Gamma Desk as a independend process and open a zmq communication channel
    This function start a new thread.

    :param int port: TCP port number to connect to (default 5998)
    :param str host: Hostname of ip address to connect to (default localhost)
    :param dict namespace: The namespace to use in the Gamma Desk console
        (Caller namespace by default)
    :param bool gui_redirect: Make the gui mapper accesible in this thread
    :param bool as_server: True -> This will be the zmq server
        (default False -> GD acts as server)

    Typical usage in a second Python process by the following command:

    >>>  from gdesk.external import channel
    >>>  channel.connect_to_gui()

    Connecting to a running GD instance at another computer

    >>>  channel.connect_to_gui(5998, 'FFYBVR-L1.ad.onsemi.com')

    """
    if namespace is None:
        namespace = sys._getframe(1).f_globals

    shell = Shell(namespace)

    if host not in ['localhost', '127.0.0.1']:
        config['image']['queue_array_shared_mem'] = False

    if port is None:
        ports = shell.get_watcher_ports()
        port = ports[-1]

    cmdclient = CommandClient(port, host)

    if as_server:
        print('Connecting as server')
        cqs = ZmqQueues()
        cqs.setup_as_server()
        cqs_config = cmdclient.send({'cmd': 'connect_zmq_process',
            'args': (cqs.to_json(),)}, timeout=5000, retries=1)
        pprint.pprint(cqs_config)

    else:
        print('Connecting as client')
        cqs_config = cmdclient.send({'cmd': 'connect_zmq_process',
            'args': ()}, timeout=5000, retries=1)
        pprint.pprint(cqs_config)
        cqs = ZmqQueues.from_json(cqs_config)
        cqs.setup_as_client(host)

    return init_gui(shell, cqs, gui_redirect, client=False)


def start_gui_as_child(namespace=None, gui_redirect=True):
    """
    Start Gamma Hawk as a child process and open a communication channel
    to a new thread in this process.

    :param dict workspace: The workspace to use in the Shell object
    :param bool gui_redirect: make the gui mapper accesible in this thread

    Typical usage in Canvas by the following command:

        from ghawk2.external.channel import start_ghawk_as_child; start_ghawk_as_child(globals())
    """
    if namespace is None:
        namespace = sys._getframe(1).f_globals

    shell = Shell(namespace)

    #CommQueues can only be send to other process by
    #spawn process inheritance
    cqs = CommQueues(Queue, process=True)

    start_gui(child=True, commqueues=cqs)

    return init_gui(shell, cqs, gui_redirect)


def start_gui(child=False, commqueues=None, deamon=False):
    if child:
        #Start the Gui as a subprocess using the multiprocessing module
        #hack the executable in sys (for embeded Python)
        sys.executable = python_executable()
        proc_config = {
            'default_perspective': 'base',
            'init_command': {'cmd': 'connect_process', 'args': ()}}

        process = Process(target=console.run_as_child,
            args=((), proc_config, {'cqs': commqueues}), daemon=deamon)
        process.start()

    else:
        os.system(f'start {python_executable()} -m ghawk2')


def init_gui(shell, commqueues, gui_redirect=True, client=True):
    name, tid = shell.new_interactive_thread(commqueues, client=client)

    if gui_redirect:
        gui.redirects[threading.get_ident()] = tid

    return tid




"""
Provide OrderedStruct and DictStruct classes.

Also, provide two utility functions: current_code_path() and force2bytes().
"""

import os, sys
import collections
from pathlib import Path

def current_code_path():
    """
    Return the path of the caller's source code file.

    If the caller is the shell (interactive scripting), None is returned.

    :return: Path object or None.

    Example:
        # works only when used from a .py file, not when used
        # in a shell (on standard input):
        from iskutil.names import current_code_path
        folder = current_code_path().parent

        # in a shell this is what happens:
        current_code_path() is None
        >>> True
    """

    # ensure that pathlib is available
    if Path is None:
        raise ImportError("Could not import pathlib; you need to use Python version v >= 3.4 or install a discrete pathlib module (e.g. pathlib2)")

    frame = sys._getframe(1)
    code_path = Path(frame.f_code.co_filename)

    if code_path.name == '<input>':
        # this code did not come from a .py file, but from standard
        # input; we don't have a valid Path to that
        return None

    return code_path.resolve()


class OrderedStruct:

    """
    Object that mimics a classic structure type for holding data and retains the order
    of the elements.

    You can think of this as an ordered DictStruct, but the name is 'OrderedStruct'
    because that is shorter ;-)

    All the methods of the dict class are supported, but they will return
    their values NOT ordered.  Instead, call _keys(), _values(), _items() and friends
    prepended with an underscore.

    Usage:
    .. code :: Python
        os = OrderedStruct(a=20)
        os.b = 'new attribute b'
        [key_value for key_value in os]
        >>> [('a', 20), ('b', 'new attribute b')]
    """

    def __init__(self, *args, **argv):
        if not argv is None:
            self.__dict__ = argv
            self.__keyorder__ = list(argv.keys())
        else:
            self.__keyorder__ = []

    def _keys(self):
        """
        Alternative for the standard dict.keys() which *does* return elements
        in order.
        """
        for key in self.__dict__.keys():
            if not key.startswith('__'):
                yield key

    def _values(self):
        """
        Alternative for the standard dict.values() which *does* return elements
        in order.
        """
        for key in self._keys():
            if not key.startswith('__'):
                yield self.__dict__[key]

    def _items(self):
        """
        Alternative for the standard dict.items() which *does* return elements
        in order.
        """
        for key in self._keys():
            if not key.startswith('__'):
                yield (key, self.__dict__[key])

    def __repr__(self):
        """
        Return string reperesentation of the OrderedStruct.
        """
        return "OrderedStruct(keys=['{}'])".format(
            "','".join(k for k in self.__keyorder__ if not k.startswith('_'))
        )

    def __str__(self):
        """
        Human-readable representation of the OrderedStruct.
        """
        s = ''
        for key in self.__keyorder__:
            s += '%s: %s\n' % (key, getattr(self, key))
        return s[:-1]

    def __setattr__(self, name, val):
        if name in ['__dict__', '__keyorder__']:
            object.__setattr__(self, name, val)
        elif not name  == '__keyorder__':
            self.__dict__[name] = val
            if not name in self.__keyorder__:
                self.__keyorder__.append(name)

    def __delattr__(self, name):
        index = self.__keyorder__.index(name)
        self.__keyorder__.pop(index)
        self.__dict__.pop(name)

    def __add__(self, other):
        """
        Return an OrderedStruct which is the superset of the current and the given
        OrderedStruct.
        """
        if not isinstance(other, OrderedStruct):
            raise ValueError("Operation '+=' (__iadd__) is only defined for other OrderedStruct instances")
        result = OrderedStruct()
        result.__keyorder__ = self.__keyorder__.copy()
        result.__dict__.update(self.__dict__)
        return result.__iadd__(other)

    def __iadd__(self, other):
        """
        Extend the current OrderedStruct by updating the current values with the values
        from the given OrderedStruct.
        """
        if not isinstance(other, OrderedStruct):
            raise ValueError("Operation '+=' (__iadd__) is only defined for other OrderedStruct instances")
        keyorderbackup = self.__keyorder__.copy()
        self.__dict__.update(other.__dict__)
        #note that this update will overwrite self.__keyorder__
        self.__keyorder__ = keyorderbackup
        for key in other.__keyorder__:
            if not key in self.__keyorder__:
                self.__keyorder__.append(key)
        return self

    def __iter__(self):
        """
        Return all keys in the OrderedDict.
        """
        for key in self.__keyorder__:
            yield key, self.__dict__[key]

    def __getitem__(self, index):
        return self.__dict__[self.__keyorder__[index]]

    def __getattr__(self, name):
        """
        Get the value of any attribute name in the internal dictionary.

        This exposes the dict methods themselves as well, such as keys(), items(), get(),
        clear(), update(), ...
.
        If the name is not present in the dictionary, raise AttributeError.
        That makes it compatible to hasattr(), which only catches
        AttributeErrors, but not KeyErrors.

        Note: this method returns 'None' for all attribute names
        which start with double underscore.  This breaks hasattr(),
        which will return True for *any* such attributename.
        """
        if name[0:2] == '__':
            return None

        return getattr(self.__dict__, name)

    def __contains__(self, key):
        """
        Check if the DictStruct contains the given key.

        The main purpose of implementing this is speed: if it is missing,
        the 'in' operator will to an iteration over all the elements instead of just
        using the hash lookup.
        """
        return key in self.__dict__
        
    def __getstate__(self):
        target = dict()
        for key, item in self.__dict__.items():
            target[key] = item
        return target
        
    def __setstate__(self, state):
        for key, item in state.items():
            self.__dict__[key] = item         


class DictStruct(object):

    """
    Object that mimics a classic structure type for holding data.

    You can consider it to be a dict with all the key/value pairs as object
    attributes, making it easier to do auto-completion on.

    Regular dict methods such as keys(), values() and get() are all present.

    Usage:
    .. code :: Python
        from iskutil.names import DictStruct
        d = DictStruct(x=50)
        d.a = 'new attribute a on first level'
        d.b = 508
        print(d.a, d.b)
    """

    def __init__(self, dictionary=None):
        '''dictionary: use the existing dictionary for this class'''
        if not dictionary is None:
            self.__dict__ = dictionary

    def __repr__(self):
        """
        Return string reperesentation of the DictStruct.
        """
        return "DictStruct(keys=['{}'])".format(
            "','".join(k for k in self.keys())
        )

    def __str__(self):
        s = ''
        for key, val in self.__dict__.items():
            s += '%s: %s\n' % (key, str(val))
        return s[:-1]

    def __getattr__(self, name):
        """
        Get the value of any attribute name in the internal dictionary.

        This exposes the dict methods themselves as well, such as keys(), items(), get(),
        clear(), update(), ...

        If the name is not present in the dictionary, raise AttributeError.
        That makes it compatible to hasattr(), which only catches
        AttributeErrors, but not KeyErrors.

        Note: this method returns 'None' for all attribute names
        which start with double underscore.  This breaks hasattr(),
        which will return True for *any* such attributename.
        """
        if name[0:2] == '__':            
            raise AttributeError(name)

        return getattr(self.__dict__, name)

    def __iadd__(self, other):
        """
        Extend the current DictStruct by updating the current values with the values
        from the given DictStruct.
        """
        if not isinstance(other, DictStruct):
            raise ValueError("Operation '+=' (__iadd__) is only defined for other DictStruct instances")
        self.__dict__.update(other.__dict__)
        return self

    def __iter__(self):
        """
        Allow to loop over all keys in the DictStruct.

        Ignore keys which start with an underscore.
        """
        keys = list(self.__dict__.keys())
        for key in keys:
            if not key.startswith('_'):
                yield key

    def _clear(self, keys):
        """
        Remove the given keys (or key) from the DictStruct.

        This is a public function, but starts with an underscore to avoid it showing
        up in auto-complete, where you expect only the *content* of the DictStruct
        to appear.

        Usage:
            ds._clear(['a', 'b', 'x'])
            ds._clear('my_var')
        """
        # we expect a list; convert a single string to a list
        if isinstance(keys, str):
            keys = [keys]
        for key in list(keys):
            del self.__dict__[key]

    def __getitem__(self, key):
        """
        Get the given item using obj[angle_bracket] syntax.

        Is essentially the same as using obj.direct_access.
        """

        return self.__dict__[key]

    def __setitem__(self, key, value):
        """
        Set the given item using obj[angle_bracket] syntax.

        Is essentially the same as using obj.direct_access.
        """
        self.__dict__[key] = value

    def __delitem__(self, key):
        """
        Remove the given item from the dictstruct.

        Is an alternative to _clear().
        """
        del self.__dict__[key]

    def __contains__(self, key):
        """
        Check if the DictStruct contains the given key.

        The main purpose of implementing this is speed: if it is missing,
        the 'in' operator will to an iteration over all the elements instead of just
        using the hash lookup.
        """
        return key in self.__dict__


def addprop(inst, name, getter, setter=None, doc=None):
    cls = type(inst)
    if not hasattr(cls, '__perinstance'):
      cls = type(cls.__name__, (cls,), {})
      cls.__perinstance = True
      inst.__class__ = cls #not sure this is needed !
    setattr(cls, name, property(getter, setter, doc))

def fromfilename(fileName):
    '''parse a filename to a valid identifier
    extension is removed
    '''
    basename = os.path.basename(fileName)
    (rootname, ext) = os.path.splitext(basename)

    rootname = rootname.replace(' ','_')
    varname = ''

    ch = rootname[0]

    if ch.isdigit():
        varname += 'x' + ch
    elif ch.isidentifier():
        varname += ch
    else:
        varname += 'x' + hex(ord(ch))

    for i in range(1, len(rootname)):
        ch = rootname[i]
        if ch.isidentifier() or ch.isdigit():
            varname += ch
        else:
            varname += hex(ord(ch))

    if not varname.isidentifier():
        raise Exception('could not parse to valid identifier')

    return varname

def find_names(obj):
    #return the list of names refering to the object
    #
    #http://pythonic.pocoo.org/2009/5/30/finding-objects-names
    import gc, sys

    #Well, since the names of all locals are known at compile-time, Python optimizes function locals,
    #putting them into an array instead of a dictionary, which speeds up access tremendously.
    #But there is a way to get at the locals in a dictionary, namely the locals() function.
    #There must be some way to get Python to create it for us!
    #This is best done accessing the frame object's f_locals attribute which creates and caches this dictionary,
    #and that can be used to create dict references to all locals if the whole chain
    #of currently executed frames is traversed, as the first for loop does.
    #
    #next lines are needed to find names of compiled objects
    #not needed for life objects
    #frame = sys._getframe()
    #for frame in iter(lambda: frame.f_back, None):
    #    frame.f_locals
    #=========

    result = []
    for referrer in gc.get_referrers(obj):
        if isinstance(referrer, dict):
            for k, v in referrer.items():
                if v is obj:
                    result.append(k)
    return result

def force2bytes(s):
    """
    Convert the given string to bytes by encoding it.

    If the given argument is not a string (already bytes?), then return it verbatim.

    :param s: Data to encode in bytes, or any other object.
    :return: Bytes-encoded variant of the string, or the given argument verbatim.
    """
    if isinstance(s, str):
        return s.encode()
    else:
        return s



# -*- coding: utf-8 -*-
import sys

from qtpy import QtGui, QtCore, QtWidgets
from .node import Node

class DictionaryTreeModel(QtCore.QAbstractItemModel):
    """Data model providing a tree of an arbitrary dictionary"""

    def __init__(self, root, parent=None):
        super(DictionaryTreeModel, self).__init__(parent)
        self._rootNode = root

    def rowCount(self, parent):
        """the number of rows is the number of children"""
        if not parent.isValid():
            parentNode = self._rootNode
        else:
            parentNode = parent.internalPointer()

        return parentNode.childCount()

    def columnCount(self, parent):
        """Number of columns is always 2 since dictionaries consist of key-value pairs"""
        return 2

    def data(self, index, role):
        """returns the data requested by the view"""
        if not index.isValid():
            return None

        node = index.internalPointer()

        if role == QtCore.Qt.DisplayRole or role == QtCore.Qt.EditRole:
            return node.data(index.column())

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        """this method gets called when the user changes data"""
        if index.isValid():
            if role == QtCore.Qt.EditRole:
                node = index.internalPointer()
                node.setData(index.column(), value)
                return True
        return False

    def headerData(self, section, orientation, role):
        """returns the name of the requested column"""
        if role == QtCore.Qt.DisplayRole:
            if section == 0:
                return "Key"
            if section == 1:
                return "Value"

    def flags(self, index):
        """everything is editable"""
        return (QtCore.Qt.ItemIsEnabled |
                QtCore.Qt.ItemIsSelectable |
                QtCore.Qt.ItemIsEditable)

    def parent(self, index):
        """returns the parent from given index"""
        node = self.getNode(index)
        parentNode = node.parent()
        if parentNode == self._rootNode:
            return QtCore.QModelIndex()

        return self.createIndex(parentNode.row(), 0, parentNode)

    def index(self, row, column, parent):
        """returns an index from given row, column and parent"""
        parentNode = self.getNode(parent)
        childItem = parentNode.child(row)

        if childItem:
            return self.createIndex(row, column, childItem)
        else:
            return QtCore.QModelIndex()

    def getNode(self, index):
        """returns a Node() from given index"""
        if index.isValid():
            node = index.internalPointer()
            if node:
                return node

        return self._rootNode

    def insertRows(self, position, rows, parent=QtCore.QModelIndex()):
        """insert rows from starting position and number given by rows"""
        parentNode = self.getNode(parent)

        self.beginInsertRows(parent, position, position + rows - 1)

        for row in range(rows):
            childCount = parentNode.childCount()
            childNode = Node("untitled" + str(childCount))
            success = parentNode.insertChild(position, childNode)

        self.endInsertRows()
        return success

    def removeRows(self, position, rows, parent=QtCore.QModelIndex()):
        """remove the rows from position to position+rows"""
        parentNode = self.getNode(parent)
        self.beginRemoveRows(parent, position, position + rows - 1)

        for row in range(rows):
            success = parentNode.removeChild(position)

        self.endRemoveRows()
        return success

    def to_dict(self):
        return self._rootNode.to_dict()


def node_structure_from_dict(datadict, parent=None, root_node=None):
    """returns a hierarchical node stucture required by the TreeModel"""
    if not parent:
        root_node = Node('Root')
        parent = root_node
        
    if isinstance(datadict, dict):
        iterator = datadict.items()
        
    elif isinstance(datadict, list):
        iterator = zip(range(len(datadict)), datadict)

    for name, data in iterator:
        node = Node(name, parent)
        if isinstance(data, (dict, list)):
            node = node_structure_from_dict(data, node, root_node)
        else:
            node.value = str(data)

    return root_node


class DictionaryTreeWidget(QtWidgets.QTreeView):
    """returns an object containing the tree of the given dictionary d.
    example:

    tree = DictionaryTree(d)
    tree.edit()
    d_edited = tree.dict()

    d_edited contains the dictionary with the edited data.
    this has to be refactored...
    """

    def __init__(self, d):
        super(DictionaryTreeWidget, self).__init__()
        self.load_dictionary(d)
    def load_dictionary(self,d):
        """load a dictionary into my tree applicatoin"""
        self._d = d
        self._nodes = node_structure_from_dict(d)
        self._model = DictionaryTreeModel(self._nodes)
        self.setModel(self._model)

    def to_dict(self):
        """returns a dictionary from the tree-data"""
        return self._model.to_dict()


class DictionaryTreeDialog(QtWidgets.QDialog):
    """guidata motivated dialog for editin dictionaries

    """

    def __init__(self, d):
        super(DictionaryTreeDialog, self).__init__()
        treeWidget = DictionaryTreeWidget(d)
        for c in range(treeWidget._model.columnCount(None)):
            treeWidget.resizeColumnToContents(c)
        self.treeWidget = treeWidget

        self.buttonOk = QtWidgets.QPushButton('Ok', self)
        self.buttonCancel = QtWidgets.QPushButton('Cancel', self)

        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(self.treeWidget)

        bhbox = QtWidgets.QHBoxLayout()
        bhbox.addStretch()
        bhbox.addWidget(self.buttonOk)
        bhbox.addWidget(self.buttonCancel)

        vbox.addLayout(bhbox)
        self.setLayout(vbox)

        self.connect(self.buttonOk, QtCore.SIGNAL('clicked()'), self.accept)
        self.connect(self.buttonCancel, QtCore.SIGNAL('clicked()'), self.closeCancel)

    def edit(self):
        return self.exec_()

    def to_dict(self):
        return self.treeWidget.to_dict()

    def closeCancel(self):
        d = self.treeWidget._d
        self.treeWidget.load_dictionary(d)
        self.reject()

    def closeEvent(self, event):
        self.closeCancel()


if __name__=='__main__':

    try:
        app = QtGui.QApplication(sys.argv)
    except:
        app = QtGui.qApp

    d = {'First name': 'Maximus',
         'Last name': 'Mustermann',
         'Nickname': 'Max',
         'Address':{ 'Street': 'Musterstr.',
                     'House number': 13,
                     'Place': 'Orthausen',
                     'Zipcode': 76123},
         'An Object': float,
         'Great-grandpa':{
             'Grandpa':{
                 'Pa': 'Child'}}
     }

    tree = DictionaryTreeDialog(d)

    if tree.edit():
        print('Accepted:')
    else:
        print('Cancelled')

    edited_dict = tree.to_dict()
    print('\nEdited dict: {}'.format(edited_dict))
    print('\nEdited dict is the same as input dict: {}'.format(edited_dict==d))
    print('\nMy object is still of type: {}'.format(edited_dict['An Object']))



import sys, os
import platform
import traceback
import time
import argparse
from pathlib import Path

from .interactive import interact
from . import workspace, manager, using, using_top, use

from .__version__ import VERSION, VERSION_INFO

modname = '.'.join(globals()['__name__'].split('.')[:-1])
PATH_SEPERATOR = ';' if sys.platform == 'win32' else ':'

HEADER = f"""Live Scripting Environment {VERSION_INFO}
==================================="""
    
epilog = f"""Searchpaths
-----------
    The list of absolute paths is available as {modname}.manager.path
    
Configuring the searchpaths:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~    

{modname } optional argument -p PATH, --path PATH:
    PATH is '{PATH_SEPERATOR}'-separated list of directories
    If not given, environment variable LIVEPATH is used.

Use of system environment variable LIVEPATH:
    '{PATH_SEPERATOR}'-separated list of directories
    If not given, search for higher directiory containing __live__.py

From withon Python:    
    {modname}.manager.path.append_path(some_path)
    
Examples
--------
          
{modname} -u script arg1 arg2

            Execute '__main__' function of 'my_script.py' script file and use 
            'arg1' as first argument.
            
{modname} -u script.hello_world

            Execute the 'hello_world' function of 'script.py' script file.         
            
{modname} arg1 arg2

            Execute '__main__' function of '__live__.py' script file and use 
            2 arguments: 'arg1' and 'arg2'
            
{modname} -i -u script.hello_world

            Enter interative mode after executing script.hello_world()        
"""

   
def argparser():    
    parser = argparse.ArgumentParser(description=HEADER, prog=f'python -m {modname}', 
        formatter_class=argparse.RawDescriptionHelpFormatter, epilog=epilog)        
        
    parser.add_argument("arguments", type=str, help="Arguments", nargs='*')
    
    parser.add_argument("-p", "--path", type=str,
        help=f"""'{PATH_SEPERATOR}'-separated list of directories prefixed to the default live search path. 
If not given, environment variable LIVEPATH is used. The list of absolute paths is available as {modname}.manager.path.""")    
        
    parser.add_argument("-u", "--using", help="Start the __main__ function from this live module", nargs=1)
    parser.add_argument("-l", "--list", help="List the available scripts", action='store_true')

    parser.add_argument("-i", "--interactive", help="Start Python interactive",  action='store_true')
    parser.add_argument("-k", "--keycomplete", help="Enable key completer in interactive mode",  action='store_true')
    
    return parser

def argexec():
    parser = argparser()   
    args = parser.parse_args()    
    
    # if len(sys.argv) <= 1:
        # parser.print_help()
    
    ws = workspace    
        
    live_paths = None
    
    if args.path:
        live_paths = args.path.split(PATH_SEPERATOR)
    elif 'LIVEPATH' in os.environ.keys():
        #This is automatic added by __init__
        pass
    else:
        print('No live search paths defined')        
        path = Path('.').absolute()
        depth = 20
        while True:
            try:
                for item in os.scandir(path):
                    if item.is_dir() and (path / item.path / '__live__.py').exists():
                        p = str(path / item.path)
                        print(f'Found {p} having __live__.py')
                        live_paths = [p]
                        break
            except:
                print(f'Could not scan {path}')
            if depth == 0 or path == path.parent:
                break
            path = path.parent
            depth -= 1

    if not live_paths is None:
        manager.path.clear()
        for path in live_paths:
            manager.append_path(path)
        
    if args.list:
        print('Available top level scripts:')
        print()
        for item in dir(using_top):
            print(f'  {item}')
        print()
               
    ws['using'] = using
    ws['use'] = use
    ws['__name__'] = '__main__'
    ws['__interactive__'] = True if args.interactive else False
    
    if args.using:
        use_name = args.using[0]
        if not '.' in use_name:
            func = eval(f'using_top.{use_name}.__main__')         
        else:
            func = eval(f'using_top.{use_name}')         
    else:
        try:
            func = eval(f'using_top.__live__.__main__')            
        except:
            func = None
            
    argv = args.arguments            

    if func is None:
        if len(sys.argv) <= 1:
            parser.print_help()        
        
    else:
        try:
            ws['__returned_value__'] = func(*argv)
        except:
            traceback.print_exc()            
      
    if ws['__interactive__'] or func is None:        
        banner = 'Type using.<TAB> to get a list of the top level scripts\n'
        banner += 'exit() to exit'
        completer = 'key' if args.keycomplete else 'standard'
        interact(ws, banner, completer=completer)
        
    return ws
        
    



import pathlib
from functools import wraps
from ... import config

if config.get('qapp', False):
    #Do not import anything from Qt if not needed !!!
    #Causes crach in Canvas !
    from qtpy import QtWidgets
    from ...dialogs import formlayout    
    from ...dialogs import base as dialogs
    
from ...core.gui_proxy import gui, GuiProxyBase, StaticGuiCall
      
class DialogGuiProxy(GuiProxyBase):    
    category = 'values'
    
    def __init__(self):
        pass
        
    def attach(self, gui):
        gui.dialog = self
        gui.question = self.question
        gui.msgbox = self.msgbox
        gui.inputdlg = self.inputdlg
        gui.getstring = self.getstring
        gui.getfile = self.getfile
        gui.getfiles = self.getfiles
        gui.getfilename = self.getfilename        
        gui.getfilenames = self.getfilenames
        gui.getpath = self.getpath
        gui.getdir = self.getdir
        gui.putfile = self.putfile
        gui.putfilename = self.putfilename
        gui.filterlist = self.filterlist       
        gui.fedit = self.fedit       
                
    @StaticGuiCall
    def msgbox(message, title = '', icon = 'none'):
        """
        Creates a message box
        
        :param message: the message
        :type message: str
        :param title: the title
        :type title: str
        :param icon: 'none', 'help, 'info', 'warn', 'error'
        :type icon: str
        """
        dialogs.messageBox(message, title, icon)         
        
    @StaticGuiCall
    def question(question='Do you press the No Button?', title='Question'):
        """Yes/No question"""
        return dialogs.questionBox(question, title)         
     
    @StaticGuiCall
    def inputdlg(prompt='', default='', title='Input', masked=False, timeout=None):
        """
        Open a input dialog. The user can enter a string.
        
        :param str prompt: The prompt
        :param str default: The default string
        :param str title: The title of the dialog
        :param bool masked: Mask the string as stars `******`
        :param timeout: After time, continue the flow. Expressed in miliseconds
        :type timeout: None or float
        :returns: The user string
        :rtype: str
        
        Timeout can not be combined with mask.        
        """
        if isinstance(prompt, str):
            if masked :
                return dialogs.getString(prompt, default, title, 'Password')
            else:
                return dialogs.getStringTimeout(prompt, default, title, timeout=timeout)
        else:
            return dialogs.getMultiString(prompt, default, title, timeout=timeout)       
                        
    @StaticGuiCall
    def getstring(prompt='', default='', title='Input', echo='Normal', timeout=None):
        """\
        Show a popup-window to ask the user some textual input.

        Makes use of QtWidgets.QInputDialog.getText; see
        https://srinikom.github.io/pyside-docs/PySide/QtGui/QInputDialog.html#PySide.QtGui.PySide.QtGui.QInputDialog.getText

        :param str prompt: The explanation that is visible just above the text input field.
        :param str default: The text that is already present in the editable input field.
        :param str title: The name of the pop-window (shown in its title bar).
        :param str echo: 'Normal' for normal text entry; 'Password' for password entry. See
          http://doc.qt.io/qt-4.8/qlineedit.html#EchoMode-enum
        """                
        if timeout is None:
            return dialogs.getString(prompt, default, title, echo=echo)
            
        else:
            return dialogs.getStringTimeout(prompt, default, title, echo=echo, timeout=timeout)    

    @StaticGuiCall
    def getfile(filter='*.*', title='open', file=None):
        """
        Show a file dialog to open a file.
        Return a string tuple of  file path and choosen filter
        
        :param str filter: a filter of the form \\*.tif
        :param str title: the window title
        :param str file: default file path
        :return: The file path and choosen filter
        :rtype: tuple(str, str)
        """
        return dialogs.getFile(filter, title, file)    
    
    @staticmethod
    def getfilename(filter='*.*', title='open', file=None):    
        return DialogGuiProxy.getfile(filter, title, file)[0]
        
    @StaticGuiCall
    def getfiles(filter='*.*', title='open', file=None):
        """
        Multi file selection version of getfile
        """
        return dialogs.getFiles(filter, title, file)  
    
    @staticmethod    
    def getfilenames(filter='*.*', title='open', file=None):
        """
        Multi file selection version of getfilename
        """
        return DialogGuiProxy.getfiles(filter, title, file)[0]
        
    @StaticGuiCall
    def selectfile(filter='*.*', title='Select', default_path=None):
        selected_files = dialogs.selectFiles(filter, title, default_path)
        if len(selected_files) == 0: return None
        return selected_files[0]        
    
    @staticmethod    
    def getpath(startpath=None, title='select a Directory'):
        """
        Show a directory dialog to choose a dir
        :return: The directory
        :rtype: pathlib.Path
        """        
        return pathlib.Path(DialogGuiProxy.getdir(startpath, title))   
        
    @StaticGuiCall
    def getdir(startpath=None, title='select a Directory'):
        """
        Show a directory dialog to choose a dir
        
        :returns: The directory
        :rtype: str
        """
        return dialogs.getMap(startpath, title)    

    @StaticGuiCall
    def putfile(filter='*.*', title='save', file=None, defaultfilter=""):
        """
        Open a save file dialog and return a file name selected by the user.
        The file does not have to exist.
                        
        :param str filter: File type filter, filters are seperated by ``;;``
            e.g.: \\*.tif, Image (\\*.tif \\*.png), Text(\\*.txt);;Image (\\*.tif \\*.png)
        :param str title: Title of the dialog
        :param file: Propose of file name
        :type file: str or None
        :param str defaultfilter: Default filter to use                
        
        :returns: The filename and used filter
        :rtype: str, str
        
        Example
        
        >>> filename, filter = gui.putfile('Text(*.txt);;Image (*.tif *.png)', "Give it a filename", r"C:\\temp\\default.tif", "Image (*.tif *.png)")
        """    
        return dialogs.putFile(filter, title, file, defaultfilter)  

    @staticmethod
    def putfilename(filter='*.*', title='save', file=None, defaultfilter=""):
        """       
        Open a save file dialog and return a file name selected by the user.
        The file does not have to exist.
                        
        :param str filter: File type filter, filters are seperated by ``;;``
            e.g.: \\*.tif, Image (\\*.tif \\*.png), Text(\\*.txt);;Image (\\*.tif \\*.png)
        :param str title: Title of the dialog
        :param file: Propose of file name
        :type file: str or None
        :param str defaultfilter: Default filter to use
        
        :returns: The filename
        :rtype: str
        
        Example
        
        >>> filename = gui.putfile('Text(*.txt);;Image (*.tif *.png)', "Give it a filename", r"C:\\temp\\default.tif", "Image (*.tif *.png)")
        """
        return DialogGuiProxy.putfile(filter, title, file, defaultfilter)[0]          

    @StaticGuiCall
    def filterlist(items=None, selection=None, filter=None):
        """
        Open a items filter dialog.
        The user have to select items from it.
        
        selection = None : select nothing
                  = one_item : select the one item
                  = []: use checkable items
                  = [str1, str2] : check items on string value
        filter    = None: check nothing
                  = False: check everything
                  = ``\w*01\w*``: use re to match string
        """
        
        #import ghawk2.dialogs.filterlist as filterlist
        
        if items is None:
            items = list()        
        
        if isinstance(selection, list): 
            multiple = True
        else:
            multiple = False

        fl = dialogs.filterlist.FilterList(items, multiple)        
        
        if not multiple:
            if not selection is None:
                fl.selectItem(selection)
            
        else:
            if filter is False:
                fl.checkAll(True)

            elif isinstance(filter, str):
                fl.checkItemsRe(filter) 
                
            elif isinstance(selection, list):
                fl.checkItems(selection)                                       
            
        dialog_code = fl.exec_()
        
        if dialog_code == QtWidgets.QDialog.DialogCode.Rejected:
            return None
        
        if multiple:
            return fl.checkedItems()
        else:
            return fl.selectedItem()        
            
            
    # TO DO: formlayout.fedit doc string can not be available if the process
    #        is not allowed to load QT (like Canvas GUI)
    @StaticGuiCall
    # @wraps(formlayout.fedit)
    def fedit(*args, **kwargs):
        """
        Create form dialog and return result
        (if Cancel button is pressed, return None)

        :param tuple data: datalist, datagroup (see below)
        :param str title: form title
        :param str comment: header comment
        :param QIcon icon: dialog box icon
        :param QWidget parent: parent widget
        :param str ok: customized ok button label
        :param str cancel: customized cancel button label
        :param tuple apply: (label, function) customized button label and callback
        :param function apply: function taking two arguments (result, widgets)
        :param str result: result serialization ('list', 'dict', 'OrderedDict',
                                                 'JSON' or 'XML')
        :param str outfile: write result to the file outfile.[py|json|xml]
        :param str type: layout type ('form' or 'questions')
        :param bool scrollbar: vertical scrollbar
        :param str background_color: color of the background
        :param str widget_color: color of the widgets

        :return: Serialized result (data type depends on `result` parameter)
        
        datalist: list/tuple of (field_name, field_value)
        datagroup: list/tuple of (datalist *or* datagroup, title, comment)
        
        Tips:
          * one field for each member of a datalist
          * one tab for each member of a top-level datagroup
          * one page (of a multipage widget, each page can be selected with a 
            combo box) for each member of a datagroup inside a datagroup
           
        Supported types for field_value:
          - int, float, str, unicode, bool
          - colors: in Qt-compatible text form, i.e. in hex format or name (red,...)
                    (automatically detected from a string)
          - list/tuple:
              * the first element will be the selected index (or value)
              * the other elements can be couples (key, value) or only values
        """    
        return formlayout.fedit(*args, **kwargs)         
                        



"""A collections of utils."""
import sys

import numpy as np


def new_id_using_keys(keys):
    """Find the lowest new id starting from already exisiting ids."""
    keys = np.array(keys)

    if len(keys) == 0:
        return 1

    keys.sort()
    np.diff(keys)
    gaps = np.where(np.diff(keys) > 1)[0]

    if len(gaps) > 0:
        key = keys[gaps[0]] + 1
    else:
        key = keys[-1] + 1

    return key


def lazyf(template):
    """Do a f-string formating."""
    frame = sys._getframe(1)
    result = eval('f"""' + template + '"""', frame.f_globals, frame.f_locals)
    return result


def clip_values(dtype):
    """Get the lowest and highest clipvalue for a certain numpy data type."""
    if dtype == 'uint8':
        low, high = 0, 255

    elif dtype == 'uint16':
        low, high = 0, 65535

    elif dtype in ['float', 'double']:
        low, high = 0, 1

    else:
        raise AttributeError(f'clip values not defined for {dtype}')

    return low, high


def clip_array(array, dtype):
    """Clip the array to clipvalues of dtype."""
    if dtype in ['uint8', 'uint16']:
        array = array.clip(*clip_values(dtype)).astype(dtype)

    return array



import pathlib
import numpy as np 

from qtpy import QtCore, QtGui, QtWidgets

from ... import config, gui

from ...graphics.view import SceneView
from ...graphics.items import createCurve, Indicator
from ...graphics.rulers import TickedRuler, Grid
from ...graphics.point import Point
from ..base import BasePanel, CheckMenu

from ..imgview import fasthist

respath = pathlib.Path(config['respath'])

colors = {
    'K': QtGui.QColor(0, 0, 0),
    'RK': QtGui.QColor(255, 0, 0),
    'R': QtGui.QColor(255, 0, 0),
    'G': QtGui.QColor(0, 255, 0),
    'B': QtGui.QColor(0, 0, 255),
    'RR': QtGui.QColor(192, 0, 0),
    'RG': QtGui.QColor(64, 127, 0),
    'RB': QtGui.QColor(64, 0, 127)}    

class LevelPlot(QtWidgets.QWidget):
    
    blackPointChanged = QtCore.Signal(object)
    whitePointChanged = QtCore.Signal(object)
    
    def __init__(self, parent):
        super().__init__(parent=parent)
        
        self.scene = QtWidgets.QGraphicsScene()
        self.view = SceneView(self)  
        self.view.freeze_y0 = True
        
        self.view.scale = [100, -100]
        self.view.setScene(self.scene)
        
        #self.view.fixScaleY = True
        
        self.vbox = QtWidgets.QVBoxLayout()
        self.vbox.setContentsMargins(0,0,0,0)
        self.vbox.addWidget(self.view)
        self.setLayout(self.vbox)
        
        self.create_x_ruler()
        self.create_y_ruler()
        # self.create_grid()       
        
        self.indicators = []
        
        self.create_indicator(0, QtGui.QColor(0,0,127), 'B:%0.5g', self.IndicatorBMoved)                        
        self.create_indicator(256, QtGui.QColor(255,127,127), 'W:%0.5g', self.IndicatorWMoved)        
        
        self.curves = dict()                      
        self.plot_curve('K', [0, 256], [0, 0])
        #self.plot_curve('G', [-2, -1, 0 , 1, 2], [0, 0.5, 1.0, 0.5 ,0])
        #self.plot_curve('B', [-2, -1, 0 , 1, 2], [1, 1.0, 1.0, 0.25 ,0])
        
        for indicator in self.indicators:
            indicator.attach_curves(self.curves)
        
        self.view.translated.connect(self.update_rulers)
        self.view.scaled.connect(self.update_rulers)
        self.view.scale_ended.connect(self.update_rulers)   
        self.view.zoom_full.connect(self.zoomFull)   
        
        self.xmin = 0
        self.xmax = 1e6
        self.ymin = 0
        self.ymax = 256

    def create_indicator(self, xpos, color, text = None, slot=None):
        indicator = Indicator(color, text, parent=self.x_ruler)
        indicator.setPos(xpos, 0)
        indicator.setZValue(2.0)
        indicator.mouse_released.connect(slot)
        self.indicators.append(indicator)
        #self.scene.addItem(indicator)
        
    def create_x_ruler(self):
        x0, y0, x1, y1 = self.view.viewRectCoord()
        self.x_ruler = TickedRuler(0, x0, x1, abs(self.view.scale[0]), noDecimals=False, parent=None)       
        self.v_grid = Grid(self.x_ruler,  parent=None)
        self.x_ruler.setPos(0, y0 - self.view.pixelSize()[1] * 22)     
        self.v_grid.setPos(0, y0 - self.view.pixelSize()[1] * 22)     
        self.x_ruler.setZValue(1.0)
        self.v_grid.setZValue(-1)
        self.scene.addItem(self.x_ruler)              
        self.scene.addItem(self.v_grid)              
        
    def create_y_ruler(self):
        x0, y0, x1, y1 = self.view.viewRectCoord()
        self.y_ruler = TickedRuler(-90, y0, y1, abs(self.view.scale[1]), noDecimals=False)  
        self.h_grid = Grid(self.y_ruler)        
        self.y_ruler.setPos(x0 + self.view.pixelSize()[0] * 22, 0)     
        self.h_grid.setPos(x0 + self.view.pixelSize()[0] * 22, 0)     
        self.y_ruler.setZValue(0.9)
        self.h_grid.setZValue(-1)
        self.scene.addItem(self.y_ruler)          
        self.scene.addItem(self.h_grid)          
        
    def attach_indicators(self):
        for indicator in self.indicators:
            indicator.setParentItem(self.x_ruler)
            
    def hide_stuff(self):
        self.x_ruler.hide()
        self.y_ruler.hide()
                
    def update_x_ruler(self):        
        x0, y0, x1, y1 = self.view.viewRectCoord()
        self.x_ruler.setPos(0, y0 - self.view.pixelSize()[1] * 22)
        self.x_ruler.update_labels(x0, x1, abs(self.view.scale[0]))
        self.v_grid.setPos(0, y0 - self.view.pixelSize()[1] * 22)
        self.v_grid.update_labels()
        
        self.x_ruler.show()
        #self.xmin = x0
        #self.xmax = x1     
        
    def update_y_ruler(self): 
        x0, y0, x1, y1 = self.view.viewRectCoord()
        self.y_ruler.setPos(x0 + self.view.pixelSize()[0] * 22, 0)
        self.y_ruler.update_labels(y0, y1, abs(self.view.scale[1]))
        self.h_grid.setPos(x0 + self.view.pixelSize()[0] * 22, 0)
        self.h_grid.update_labels()        
        self.y_ruler.show()  

        for indicator in self.indicators:
            indicator.updates_ylabels()         
            
        self.ymin = y0 - self.view.pixelSize()[1] * 22
        self.ymax = y1

    def update_rulers(self, x, y):  
        self.update_x_ruler()
        self.update_y_ruler()   

    def remove_all_but(self, curve_ids):
        for curveid in list(self.curves.keys()):
            if curveid in curve_ids:
                continue
            oldcurve = self.curves[curveid]
            self.scene.removeItem(oldcurve)
            self.curves.pop(curveid)
        
    def plot_curve(self, curveid=None, x=[], y=[], color = None):
        oldcolor = None
        
        if curveid in self.curves.keys():
            oldcurve = self.curves[curveid]
            oldcolor = oldcurve.pen().color()
            self.scene.removeItem(oldcurve)
            
        elif curveid is None:            
            curveid = len(self.curves)            
        
        if color is None:
            if oldcolor is None:
                color = colors[curveid]
            else:
                color = oldcolor
            
        curve = createCurve(x, y, color = color)
        curve.setZValue(0)            
        
        self.curves[curveid] = curve                     
        self.scene.addItem(curve)
        
        return curveid
        
    def zoomFull(self, enforce_ymin=None):
        self.xmin = self.xmax = None
        self.ymin = self.ymax = None
        
        for curve in self.curves.values():
            xmin_cand = min(curve.xvector)
            xmax_cand = max(curve.xvector)
            ymin_cand = min(curve.yvector)
            ymax_cand = max(curve.yvector)            
            if (self.xmin is None) or (xmin_cand < self.xmin):
                self.xmin = xmin_cand            
            if (self.xmax is None) or (xmax_cand > self.xmax):
                self.xmax = xmax_cand
            if (self.ymin is None) or (ymin_cand < self.ymin):
                self.ymin = ymin_cand            
            if (self.ymax is None) or (ymax_cand > self.ymax):
                self.ymax = ymax_cand
                        
        self.ymin = enforce_ymin if not enforce_ymin is None else self.ymin
        
        if self.xmin == self.xmax:
            self.xmin -= 0.5
            self.xmin += 0.5
            
        if self.ymin == self.ymax:
            self.ymin -= 0.5
            self.ymax += 0.5
                    
        self.view.setXLimits(self.xmin, self.xmax, 22, 0)
        self.view.setYLimits(self.ymin, self.ymax, 22, 0)
        #self.resetIndicators(self.xmin, self.xmax)
        self.update_rulers(True, True)
        
    def zoomFitYRange(self, ymin=None):
        self.ymin = self.ymax = None

        for curve in self.curves.values():
            ymin_cand = min(curve.yvector)
            ymax_cand = max(curve.yvector)            
            if (self.ymin is None) or (ymin_cand < self.ymin):
                self.ymin = ymin_cand
            if (self.ymax is None) or (ymax_cand > self.ymax):
                self.ymax = ymax_cand  
                
        self.ymin = ymin if not ymin is None else self.ymin

        if self.ymin == self.ymax:
            self.ymin -= 0.5
            self.ymax += 0.5  

        self.view.setYLimits(self.ymin, self.ymax, 22, 0)
        self.update_rulers(True, True)
        
    def zoomBetweenIndicators(self):
        x0 = self.indicators[0].pos().x()
        x1 = self.indicators[1].pos().x()
        self.view.setXLimits(x0, x1, 22, 0)
        self.update_rulers(True, True)                           
        
    def resetIndicators(self, xmin=0, xmax=1):
        for indicator in self.indicators:
            x = indicator.pos().x()
            x = min(max(x, xmin), xmax)
            indicator.setPos(x, 0)    
                        
    def IndicatorBMoved(self, val):
        self.blackPointChanged.emit(val)
        
    def IndicatorWMoved(self, val):
        self.whitePointChanged.emit(val)     

    def resizeEvent(self, ev):
        self.view.setYLimits(self.ymin, self.ymax, 22, 0)
        self.update_rulers(True, True)         
        
       
class Levels(QtWidgets.QWidget):    
        
    def __init__(self, parent=None):         
        self.panel = parent
        super().__init__(parent=parent)                
        self.initUI()
        
    def initUI(self):                            
        self.setMinimumHeight(80)
        self.setMinimumWidth(200)    

        self.hbox = QtWidgets.QHBoxLayout()         
        self.hbox.addStretch()                
                
        self.levelplot = LevelPlot(self)                                         
                         
        self.vbox = QtWidgets.QVBoxLayout()
        self.vbox.setContentsMargins(0,0,0,0)
        self.vbox.setSpacing(0)
        self.setLayout(self.vbox)        
        self.vbox.addLayout(self.hbox)                
        self.vbox.addWidget(self.levelplot)                                                 

    #@staticmethod
    def image_panel(self, panel_id=None):
        if (panel_id is None) or (panel_id==False):        
            imagePanel = self.panel.bindedPanel('image')
            if imagePanel is None:
                imagePanel = gui.qapp.panels.selected('image')            
        else:            
            imagePanel = gui.qapp.panels['image'][panel_id] 
        return imagePanel                              
       
    @staticmethod       
    def xy_as_steps(xvector, yvector, stepwidth):
        yvector = yvector.reshape((yvector.shape[0], 1)).dot(np.ones((1, 2))).reshape(yvector.shape[0]*2)            
        xvector = xvector.reshape((xvector.shape[0], 1)).dot(np.ones((1, 2)))
        xvector[:,1] += stepwidth
        xvector = xvector.flatten()
        return xvector, yvector  
        
    def updateActiveHist(self):
        if self.panel.cached:
            self.updatedCachedHistogram(None)            
        else:
            self.updateHist(None)
        
        self.levelplot.zoomFitYRange(ymin=0)
            
    def updateHistOfPanel(self, panelId=None):
        if self.panel.cached:
            self.updatedCachedHistogram(panelId)            
        else:
            self.updateHist(panelId)        
            
        self.levelplot.zoomFitYRange(ymin=0)

    def updatedCachedHistogram(self, panid):      
        image_panel = self.image_panel(panid)
        chanstats = image_panel.imviewer.imgdata.chanstats 
        
        if self.panel.histSizePolicy == 'bins':
            bins = int(self.panel.histSize)               
        else:
            bins = None
            stepsize = int(self.panel.histSize)
        
        if self.panel.roi and image_panel.imviewer.roi.isVisible():
            clr_filter = set(('RK','RR', 'RG', 'RB'))
        else:
            clr_filter = set(('K', 'R', 'G', 'B'))
        
        clr_to_draw = clr_filter.intersection(set(chanstats.keys()))
        self.levelplot.remove_all_but(clr_to_draw)
        
        for clr in clr_to_draw: 
            chanstat = chanstats[clr]
            if chanstat.arr2d is None: continue
            step = stepsize if bins is None else chanstat.step_for_bins(bins)            
            hist = chanstat.histogram(step)
            if self.panel.sqrt:
                hist = hist ** 0.5              
            starts = chanstat.starts(step)            
            starts, hist = self.xy_as_steps(starts, hist, chanstat.stepsize(step))
            self.levelplot.plot_curve(clr, starts, hist) 
        
    def updateHist(self, panelId=None):
        qapp = gui.qapp
        with qapp.waitCursor():        
            use_numba = config['levels']['numba']
            relative = config['levels']['relative']
            hist2d = fasthist.hist2d                
            
            imagePanel = self.image_panel(panid)                   
                
            do_roi = self.panel.roi and imagePanel.imviewer.roi.isVisible()            

            if do_roi: 
                slices = imagePanel.imviewer.imgdata.selroi.getslices()
                arr = imagePanel.ndarray[slices]
            else:
                arr = imagePanel.ndarray    
                
            if len(arr.shape) == 2:        
                if self.panel.histSizePolicy == 'bins':
                    stepcount = int(self.panel.histSize)
                    hist, starts, stepsize = hist2d(arr, bins=stepcount-1, plot=False, use_numba=use_numba)
                    if self.panel.sqrt:
                        hist = hist ** 0.5
                    self.stepsize.setText(str(stepsize))
                    
                elif self.panel.histSizePolicy == 'step':
                    stepsize = int(self.panel.histSize)
                    hist, starts, stepsize = hist2d(arr, step=stepsize, pow2snap=True, plot=False, use_numba=use_numba)
                    if self.panel.sqrt:
                        hist = hist ** 0.5                
                    self.stepcount.setText(str(len(hist)))        
                
                if relative:
                    hist = hist /  (arr.shape[0] * arr.shape[1] * stepsize)
                    
                starts, hist = self.xy_as_steps(starts, hist, stepsize)
                
                if do_roi:
                    self.levelplot.remove_all_but(['RK'])
                    self.levelplot.plot_curve('RK', starts, hist)            
                else:
                    self.levelplot.remove_all_but(['K'])
                    self.levelplot.plot_curve('K', starts, hist)                            
                
            elif len(arr.shape) == 3 and do_roi:
            
                self.levelplot.remove_all_but(['RR','RG','RB'])
                for clr_ch, clr_str in [(0,'RR'),(1,'RG'),(2,'RB')]:        
                    if self.panel.histSizePolicy == 'bins':
                        stepcount = int(self.panel.histSize)
                        hist, starts, stepsize = hist2d(arr[:,:,clr_ch], bins=stepcount-1, plot=False, use_numba=use_numba)     
                        if self.panel.sqrt:
                            hist = hist ** 0.5
                        self.stepsize.setText(str(stepsize))                    
                        
                    elif self.panel.histSizePolicy == 'step':
                        stepsize = int(self.panel.histSize)
                        hist, starts, stepsize = hist2d(arr[:,:,clr_ch], step=stepsize, pow2snap=True, plot=False, use_numba=use_numba)            
                        if self.panel.sqrt:
                            hist = hist ** 0.5                
                        self.stepcount.setText(str(len(hist)))
                        
                    if relative:
                        hist = hist /  (arr.shape[0] * arr.shape[1] * stepsize)
                        
                    starts, hist = self.xy_as_steps(starts, hist, stepsize)                 
                    self.levelplot.plot_curve(clr_str, starts, hist) 
                    
            elif len(arr.shape) == 3 and not do_roi:     
            
                self.levelplot.remove_all_but(['R','G','B'])
                for clr_ch, clr_str in [(0,'R'),(1,'G'),(2,'B')]:        
                    if self.step_priority == 'stepcount':
                        stepcount = eval(self.stepcount.text())
                        hist, starts, stepsize = hist2d(arr[:,:,clr_ch], bins=stepcount-1, plot=False, use_numba=use_numba)     
                        if self.panel.sqrt:
                            hist = hist ** 0.5
                        self.stepsize.setText(str(stepsize))                    
                        
                    elif self.step_priority == 'stepsize':
                        stepsize = eval(self.stepsize.text())
                        hist, starts, stepsize = hist2d(arr[:,:,clr_ch], step=stepsize, pow2snap=True, plot=False, use_numba=use_numba)            
                        if self.panel.sqrt:
                            hist = hist ** 0.5                
                        self.stepcount.setText(str(len(hist)))
                        
                    if relative:
                        hist = hist /  (arr.shape[0] * arr.shape[1] * stepsize)                    
                        
                    starts, hist = self.xy_as_steps(starts, hist, stepsize)                 
                    self.levelplot.plot_curve(clr_str, starts, hist)  
        

    def updateIndicators(self, image_panel_id):
        qapp = gui.qapp
        
        if (image_panel_id is None) or (image_panel_id==False):            
            imagePanel =  qapp.panels.selected('image')
            
        else:            
            imagePanel = qapp.panels['image'][image_panel_id]
        
        ind = self.levelplot.indicators[0]
        ind.setPos(imagePanel.offset, 0)
        ind.label.updateText(ind.text % imagePanel.offset)
        ind.updates_ylabels()
        
        ind = self.levelplot.indicators[1]
        ind.setPos(imagePanel.white, 0)
        ind.label.updateText(ind.text % imagePanel.white)
        ind.updates_ylabels()        
            
    def indicZoom(self):
        self.levelplot.zoomBetweenIndicators()
        
    def fullZoom(self):
        self.levelplot.zoomFull(enforce_ymin=0)        

    def zoomFitYRange(self):
        self.levelplot.zoomFitYRange(ymin=0)           
        
    # def focusInEvent(self, event):
        # self.parent().select()     

class LevelsToolBar(QtWidgets.QToolBar):
    def __init__(self, *args, **kwargs):    
        super().__init__(*args, **kwargs) 
        self.initUi()

    @property
    def panel(self):
        return self.parent()       
        
    @property
    def levels(self):
        return self.parent().levels
        
    def initUi(self):
        self.addAction(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'update.png')), 'Refresh', self.levels.updateActiveHist)
        
        self.histSizePolicyBox = QtWidgets.QComboBox()
        self.histSizePolicyBox.addItem('bins')
        self.histSizePolicyBox.addItem('step')
        self.histSizePolicyBox.currentTextChanged.connect(self.histSizePolicyChanged)
        self.addWidget(self.histSizePolicyBox)        
        
        self.stepcount = QtWidgets.QLineEdit('64', self)
        self.stepcount.setMaximumWidth(100)
        self.stepcount.textChanged.connect(self.histSizeChanged)
        self.addWidget(self.stepcount)
        
        self.addAction(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'unmark_to_download.png')), 'Apply default offset, gain and gamma', self.panel.gain1)  
        
        self.gainSigmaMenu = QtWidgets.QMenu('Contrast')
        actGainSigma1 = QtWidgets.QAction('Gain to Sigma 1', self, triggered=lambda: self.panel.autoContrast(1))
        actGainSigma1.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'contrast_high.png')))        
        self.gainSigmaMenu.addAction(actGainSigma1)    
        actGainSigma2 = QtWidgets.QAction('Gain to Sigma 2', self, triggered=lambda: self.panel.autoContrast(2))
        actGainSigma2.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'contrast.png')))        
        self.gainSigmaMenu.addAction(actGainSigma2)    
        actGainSigma3 = QtWidgets.QAction('Gain to Sigma 3', self, triggered=lambda: self.panel.autoContrast(3))
        actGainSigma3.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'contrast_low.png')))        
        self.gainSigmaMenu.addAction(actGainSigma3)                         
        self.autoBtn = QtWidgets.QToolButton(self)
        self.autoBtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'contrast.png')))
        self.autoBtn.setMenu(self.gainSigmaMenu)
        self.autoBtn.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup)
        self.autoBtn.clicked.connect(lambda: self.panel.autoContrast())        
        self.addWidget(self.autoBtn)
        
        self.useRoiBtn = QtWidgets.QToolButton(self)
        self.useRoiBtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'region_of_interest.png')))
        self.useRoiBtn.setCheckable(True)
        self.useRoiBtn.clicked.connect(self.toggleRoi)
        self.addWidget(self.useRoiBtn)     
        
        self.sqrtBtn = QtWidgets.QToolButton(self)
        self.sqrtBtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'square_root.png')))
        self.sqrtBtn.setCheckable(True)
        self.sqrtBtn.clicked.connect(self.toggleSqrt)
        self.addWidget(self.sqrtBtn)         
        
        self.addAction(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'zoom_fit.png')), 'Zoom to full histogram', self.levels.fullZoom)        
        self.addAction(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'zoom_actual_equal.png')), 'Zoom Fit Y range', self.levels.zoomFitYRange)
        self.addAction(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'zoom_cursors.png')), 'Zoom to black white indicators', self.levels.indicZoom)        
        self.addAction(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'dopplr.png')), 'Choose colormap', self.colorMap)        
        
        fontHeight = self.fontMetrics().height()
        self.setIconSize(QtCore.QSize(fontHeight * 3 / 2, fontHeight * 3 / 2))
        
    def histSizePolicyChanged(self, text):
        self.stepcount.setText(str(self.panel.histSizes[text]))
        
    def histSizeChanged(self, text):
        histSizePolicy = self.histSizePolicyBox.currentText()
        self.panel.histSizePolicy = histSizePolicy
        self.panel.histSizes[histSizePolicy] = text
        
    def toggleRoi(self):
        sender = self.sender()
        self.panel.roi = sender.isChecked()
        self.levels.updateActiveHist()
        
    def toggleSqrt(self):
        sender = self.sender()
        self.panel.sqrt = sender.isChecked()
        self.levels.updateActiveHist()
        
    def updateButtonStates(self):
        self.histSizePolicyBox.setCurrentText(self.panel.histSizePolicy)
        self.useRoiBtn.setChecked(self.panel.roi)
        self.sqrtBtn.setChecked(self.panel.sqrt)
        
    def colorMap(self):
        panids = self.panel.panIdsOfBounded('image')
        for panid in panids:
            gui.menu_trigger('image', panid, ['View','Colormap...'])
        
        
class LevelsPanel(BasePanel):

    panelCategory = 'levels'
    panelShortName = 'basic'
    userVisible = True
    
    offsetGainChanged = QtCore.Signal(object, object, object)    
    blackWhiteChanged = QtCore.Signal(object, object)    
    
    classIconFile = str(respath / 'icons' / 'px16' / 'color_adjustment.png')

    def __init__(self, parent, panid):    
        super().__init__(parent, panid, 'levels')  

        self.histSizePolicy = config['levels'].get('hist_size_policy', 'bins')
        bins = config['levels'].get('bins', 64)
        step = config['levels'].get('step', 4)
        self.histSizes = {'bins': bins, 'step': step}
        
        self.cached = True
        self.sqrt = False
        self.roi = False
        self.sigma = 1
        
        self.levels = Levels(self)
        self.setCentralWidget(self.levels)
        
        self.toolbar = LevelsToolBar(self)
        self.addToolBar(self.toolbar)        
        
        self.fileMenu = self.menuBar().addMenu("&File")
        self.modeMenu = CheckMenu("&Mode", self.menuBar())
        self.menuBar().addMenu(self.modeMenu)        
        
        self.addMenuItem(self.fileMenu, 'Close', self.close_panel,
            statusTip="Close this levels panel",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'cross.png')))
        
        self.addMenuItem(self.modeMenu, 'Cached', self.toggle_cached, checkcall=lambda: self.cached)
        self.addMenuItem(self.modeMenu, 'Roi', self.toggle_roi, checkcall=lambda: self.roi)
        self.addMenuItem(self.modeMenu, 'Sqrt', self.toggle_sqrt, checkcall=lambda: self.sqrt)
        
        self.addBaseMenu(['image'])
        
        self.levels.levelplot.blackPointChanged.connect(self.updateBlackPoint)
        self.levels.levelplot.whitePointChanged.connect(self.updateWhitePoint)
        
        self.levels.show()
        self.toolbar.updateButtonStates()
        
        self.statusBar().hide()
        
    @property
    def histSize(self):
        return self.histSizes[self.histSizePolicy]
        
    def addBindingTo(self, category, panid):
        targetPanel = super().addBindingTo(category, panid)
        if targetPanel is None: return None
        self.offsetGainChanged.connect(targetPanel.changeOffsetGain)
        self.blackWhiteChanged.connect(targetPanel.changeBlackWhite)
        return targetPanel
        
    def removeBindingTo(self, category, panid):
        targetPanel = super().removeBindingTo(category, panid)
        if targetPanel is None: return None
        self.offsetGainChanged.disconnect(targetPanel.changeOffsetGain)        
        self.blackWhiteChanged.disconnect(targetPanel.changeBlackWhite)        
        return targetPanel         
        
    def toggle_cached(self):
        self.cached = not self.cached
        self.toolbar.updateButtonStates()
        self.levels.updateActiveHist()
        
    def toggle_roi(self):
        self.roi = not self.roi
        self.toolbar.updateButtonStates()
        self.levels.updateActiveHist()        

    def toggle_sqrt(self):
        self.sqrt = not self.sqrt
        self.toolbar.updateButtonStates()
        self.levels.updateActiveHist()          
        
    def autoContrast(self, sigma=None):
        if not sigma is None:
            self.sigma = sigma
            
        panids = self.panIdsOfBounded('image')
        for panid in panids:
            #gui.menu_trigger('image', panid, ['View', 'Gain to Sigma', f'Gain to Sigma {self.sigma}'])
            gui.qapp.panels['image'][panid].gainToSigma(self.sigma, self.roi)
            
    def gain1(self):    
        self.offsetGainChanged.emit('default', 'default', 'default')
        
    def updateBlackPoint(self, value):
        self.blackWhiteChanged.emit(value, None)
        
    def updateWhitePoint(self, value):
        self.blackWhiteChanged.emit(None, value)        
        
    def imageContentChanged(self, image_panel_id, zoomFit=False):
        self.levels.updateHistOfPanel(image_panel_id)
        if zoomFit:
            self.levels.fullZoom()

    def imageGainChanged(self, image_panel_id):
        self.levels.updateIndicators(image_panel_id)        
        
    def roiChanged(self, image_panel_id):
        if self.roi:
            self.levels.updateHistOfPanel(image_panel_id)
        
        


import sys
import io
import os
import traceback
import queue
import ctypes
import time
import threading
import multiprocessing
import builtins
import psutil
import logging
import inspect
import cProfile
import pstats
from queue import Queue
from codeop import CommandCompiler
from datetime import timedelta
from pathlib import Path

import psutil

from .conf import config
from . import stdinout
from .stdinout import ProcessStdInput
from .gui_proxy import GuiProxy, GuiMap, gui


class SyncBreaked(Exception):
    pass
    

class QueueInterpreter(object):
    def __init__(self, shell, cqs, gui_proxy=None):
    
        self.shell = shell
        self.cqs = cqs
        
        this_process = multiprocessing.current_process()
        this_thread = threading.currentThread()
        self.thread_id = thread_id = this_thread.ident
        self.console_id = None
        
        self.enable_trace = True
        self.breakable = False
        self.break_sent = False
        self.stop = False
        self.timeit = False
        
        self.enable_profile = False
        self.profile_sortby = 'cumulative'
        
        self.control_thread = None
        
        if gui_proxy is None:        
            self.gui_proxy = GuiProxy(None, cqs.gui_call_queue, cqs.gui_return_queue)
        else:
            self.gui_proxy = gui_proxy
        
        self.register_thread(thread_id, cqs)
        
        self.interpreter = Interpreter(self.shell.wsdict, thread_id)     
        self.shell.interpreters[self.thread_id] = self
        
        callbackargs = ('process', 0, (os.getpid(), this_thread.name, threading.get_ident()))        
        self.gui_proxy._call(-2, *callbackargs)
        
        print(f'{sys.executable}')
        print(f'Python {".".join(str(v) for v in sys.version_info)}')
        print(os.getcwd())
        print(f'{this_process.name} [{os.getpid()}] {this_thread.name} [{threading.get_ident()}]')
        
    def register_thread(self, tid, cqs):        
        self.stdout = stdinout.FlushPipeStream(cqs.stdout_queue, lambda: self.gui_proxy._call_no_wait(-1))
        self.shell.stdout.route_stream(self.stdout, tid)
        self.stderr = stdinout.ErrLogStream()
        self.shell.stderr.route_stream(self.stderr, tid)
        
        ProcessStdInput.stdin_queues[tid] = cqs.stdin_queue                            
        GuiMap.gui_proxies[tid] = self.gui_proxy    

    def unregister_thread(self):
        tid = self.thread_id
        self.shell.stdout.streams.pop(tid)
        self.shell.stderr.streams.pop(tid)    
        self.shell.interpreters.pop(tid)        
        ProcessStdInput.stdin_queues.pop(tid)
        GuiMap.gui_proxies.pop(tid)
    
    @staticmethod
    def create_and_interact(shell, cqs, gui_proxy=None):
        try:
            queuedinter = QueueInterpreter(shell, cqs, gui_proxy)
            queuedinter.interact()
            
        finally:
            sys.__stdout__.write(f'End of create_and_interact\n')
            sys.__stdout__.flush()        
        
    def interact(self):
        self.control_thread = threading.Thread(target=self.control_loop, name=f'control_thread{self.thread_id}', daemon=True)
        self.control_thread.start()   
        
        self.commandLoop()
        
    def control_loop(self):
        flowcode = 1
        
        self.shell.stdout.route_stream(self.stdout)
        
        try:
            while flowcode == 1:
                flowcode = self.control()
                
        finally:
            self.shell.stdout.unregister()
            sys.__stdout__.write(f'Exiting control loop of thread {self.thread_id}\n')
            sys.__stdout__.flush() 
        
    def commandLoop(self):
        flowcode = 1
        
        try:
            while flowcode == 1:
                flowcode = self.execute()   

        finally:
            self.unregister_thread()
            sys.__stdout__.write(f'Exiting command loop of thread {self.thread_id}\n')
            sys.__stdout__.flush()            
            
        
    def system_tracer(self, *args):
        if self.enable_trace and self.stop:
            raise SyncBreaked('Breaked by system tracer')
        return None
        
    def get_current_trace(self, back=50):
        lines = ''
        for frame in self.interpreter.get_code_frames(back):        
            fi = inspect.getframeinfo(frame)
            
            file_position = '  File "%s", line %d, in %s\n' % (fi.filename, fi.lineno, fi.function)
            file_code = ''
            try:
                for line in fi.code_context:
                    file_code += line
            except:
                pass
            lines = file_position + file_code + lines
        
        return lines        
        
    def get_current_locals(self, back=5):
        lines = ''
        
        for frame in self.interpreter.get_code_frames(back):
            for key, val in frame.f_locals.items():
                if key.startswith('_'):
                    continue
                lines += f'{key}: {val}\n'
                
        return lines    

    def set_console_mode(self, mode):
        gui.console.set_mode(mode, self.console_id)        
        
    def control(self):
        cqs = self.cqs
        interpreter = self.interpreter
        gui_proxy = self.gui_proxy    
        
        while True:
            try:
                mode, args, callback = cqs.cmd_queue.get()
                break
            
            except KeyboardInterrupt:
                pass          
                
        if mode == 'flow':
            cmd, *cargs = args
            
            if cmd == 'heartbeat':
                callbackargs = (mode, 0, 'heartbeat')
                retvalue = 1
            
            elif cmd == 'set_tracing':
                self.enable_trace = cargs[0]
                callbackargs = (mode, 0, f'Enable Trace: {self.enable_trace}')
                retvalue = 1
                
            elif cmd == 'set_timeit':
                self.timeit = cargs[0]
                callbackargs = (mode, 0, f'Enable Time It: {self.timeit}')
                retvalue = 1
                
            elif cmd == 'enable_profiling':
                self.enable_profile = True
                callbackargs = (mode, 0, f'Enable profiling')
                retvalue = 1
            
            elif cmd == 'trace':
                print(self.get_current_trace())
                callbackargs = (mode, 0, 'Trace printed')
                retvalue = 1
                
            elif cmd == 'locals':
                print(self.get_current_locals())
                callbackargs = (mode, 0, 'Locals printed')
                retvalue = 1

            elif cmd == 'sync_break':
                if not threading.currentThread().ident == self.thread_id and self.enable_trace:
                    self.stop = True
                    callbackargs = (mode, 0, 'Thread is asked to stop')
                else:
                    callbackargs = (mode, 1, 'Tracing is not enabled')
                retvalue = 1

            elif cmd == 'KeyboardInterrupt':
                if not threading.currentThread().ident == self.thread_id and self.breakable:
                    self.break_sent = True
                    self.interpreter.async_break()                    
                    callbackargs = (mode, 0, 'KeyboardInterrupt send')
                else:
                    callbackargs = (mode, 1, 'KeyboardInterrupt not send')
                retvalue = 1

            elif cmd == 'system_exit':
                self.interpreter.async_system_exit()
                callbackargs = (mode, 0, 'SystemExit send')
                retvalue = 0
                
            elif cmd == 'kill':                
                parent = psutil.Process(os.getpid())
                for child in parent.children(recursive=True):
                    child.kill()
                parent.kill()                
            
            elif cmd == 'finish':
                callbackargs = (mode, 0, 'Finishing')
                retvalue = 0

            else:
                callbackargs = (mode, 1, 'Command unkown')
                retvalue = 1                              
                
        elif mode == 'flow_func':
            func = args[0]
            args = args[1]
            if isinstance(func, str):   
                print('str is not supported as function pointer')  
                callbackargs = (mode, 1, 'str is not supported as function pointer')
                retvalue = 1

            elif isinstance(func, tuple):            
                try:                    
                    func = gui_proxy.decode_func(func)
                    result = func(*args)
                    error_code = 0
                    
                except Exception:
                    traceback.print_exc() 
                    result = None
                    error_code = 1
                    
                callbackargs = (mode, error_code, result)
                retvalue = 1                   
            
        elif mode == 'eval':
            try:                    
                result = [eval(arg, self.shell.wsdict) for arg in args]
                error_code = 0
                
            except Exception:
                traceback.print_exc()
                result = None
                error_code = 1
                
            callbackargs = (mode, error_code, result)
            retvalue = 1            
                
        else:        
            callbackargs = (mode, 1, 'unkown')
            retvalue = 1
            
        if callback is None:
            #There always should be a callback: {callbackargs}')
            #This is only used for the evaluate
            cqs.return_queue.put(callbackargs)
        else:
            gui_proxy._call(callback, *callbackargs)                                                                                    

        return retvalue          
        
    def execute(self):    
        cqs = self.cqs
        interpreter = self.interpreter
        gui_proxy = self.gui_proxy    
        
        while True:
            try:
                content = cqs.stdin_queue.get(timeout=3)
                mode, args, callback = content
                break
            
            except queue.Empty:
                pass
            
            except KeyboardInterrupt:
                pass

        if mode == 'interprete':
            error_code = None
            result = None
            
            try:
                self.breakable = True
                self.stop = False
                
                #the tracer seemed to be disabled after every sync break ???
                if self.enable_trace:
                    sys.settrace(self.system_tracer)
                else:
                    sys.settrace(None)
                    
                if self.enable_profile:
                    self.profile = cProfile.Profile()
                    self.profile.enable()
                
                start_moment = end_moment = time.perf_counter()
                redbull_timeout = config['console']['redbull']
                if redbull_timeout > 0:
                    self.gui_proxy.redbull.enable(redbull_timeout)
                    
                self.set_console_mode('running')
                error_code, result = interpreter.use_one_command(*args)  
                
                if self.enable_profile:
                    self.profile.disable()
                    self.enable_profile = False                    
                    profile_stats = pstats.Stats(self.profile).sort_stats(self.profile_sortby)
                    profile_stats.print_stats()

                sys.settrace(None)                
                
                if self.break_sent:
                    #A async KeyboardInterrupt was sent but still have to occur
                    #This situation is rare, but can happen
                    #Keep on stepping in the Python interpreter
                    print('WARNING: KeyboardInterrupt is on its way')
                    for i in range(100):
                        time.sleep(0.010)                                   
                
            except KeyboardInterrupt:                        
                error_code = 3
                result = 'Thread Interrupted by KeyboardInterrupt'
                print(result)                
                
            except SyncBreaked:
                error_code = 3
                result = 'Thread Interrupted by SyncBreaked'
                print(result)
                
            except Exception as ex:
                error_code = 4
                result = repr(ex)                
                
            finally:
                #Finish the side thread
                self.breakable = False
                self.break_sent = False                    
                self.gui_proxy.redbull.disable()
                
                if self.timeit:
                    end_moment =  time.perf_counter()                 
                    print(f'Elapased time {end_moment-start_moment} s')                                
                    
                callbackargs = (mode, error_code, result)
                retvalue = 1                
            
        elif mode == 'exit':
            self.unregister_thread()
            callbackargs = (mode, 0, 'Exiting')
            retvalue = 0
            
        elif mode == 'console id':
            self.console_id = args[0]
            callbackargs = (mode, 0, f'Console id set: {self.console_id}')
            retvalue = 1               

        elif mode == 'func':
            func = args[0]
            args = args[1]
            if isinstance(func, str):   
                print('str is not supported as function pointer')  
                callbackargs = (mode, 1, 'str is not supported as function pointer')
                retvalue = 1

            elif isinstance(func, tuple):            
                try:                    
                    func = gui_proxy.decode_func(func)
                    result = func(*args)
                    error_code = 0
                    
                except Exception:
                    traceback.print_exc() 
                    result = None
                    error_code = 1
                    
                callbackargs = (mode, error_code, result)
                retvalue = 1            

        else:        
            callbackargs = (mode, 1, 'unkown')
            retvalue = 1
            
        if callback is None:
            #There always should be a callback: {callbackargs}')
            #This is only used for the evaluate
            cqs.return_queue.put(callbackargs)
        else:
            gui_proxy._call(callback, *callbackargs)                                                                                    

        return retvalue  
        

class Interpreter(object):
    def __init__(self, workspace=None, thread_id=None):
        if workspace is None:
            workspace = dict()
            
        self.workspace = workspace
        self.compile = CommandCompiler()  
        self.thread_id = thread_id
        
    def compile_source(self, source, filename="<input>", symbol="auto"):
        """
        Compile and run some source in the interpreter.

        Arguments are as for compile_command().

        One several things can happen:

        1) The input is incorrect; compile_command() raised an
        exception (SyntaxError or OverflowError).  A syntax traceback
        will be printed by calling the showsyntaxerror() method.

        2) The input is incomplete, and more input is required;
        compile_command() returned None.  Nothing happens.

        3) The input is complete; compile_command() returned a code
        object.  The code is executed by calling self.runcode() (which
        also handles run-time exceptions, except for SystemExit).

        The return value is True in case 2, False in the other cases (unless
        an exception is raised).  The return value can be used to
        decide whether to use sys.ps1 or sys.ps2 to prompt the next
        line.
        """    
        if symbol == 'auto':
            if source.count('\n') == 0:  
                symbol = 'single'
            else:
                symbol = 'exec'

        #Compile the code, show syntax error
        try:
            code = self.compile(source, filename, symbol)
        except (OverflowError, SyntaxError, ValueError):
            # Case 1
            self.showsyntaxerror(filename)
            return 1, None
            
        if code is None:
            # Case 2
            # more code expected, nothing to do
            return 2, None                  

        return 0, code
        
    def exec_code(self, code):
        """
        Execute a code object.

        When an exception occurs, self.showtraceback() is called to
        display a traceback.  All exceptions are caught except
        SystemExit, which is reraised.

        A note about KeyboardInterrupt: this exception may occur
        elsewhere in this code, and may not always be caught.  The
        caller should be prepared to deal with it.
        """
        try:
            return exec(code, self.workspace)
        except (SystemExit, KeyboardInterrupt, SyncBreaked):            
            raise
        except:           
            self.showtraceback()       

    def eval_expression(self, expression):
        """
        Execute a code object.

        When an exception occurs, self.showtraceback() is called to
        display a traceback.  All exceptions are caught except
        SystemExit, which is reraised.

        A note about KeyboardInterrupt: this exception may occur
        elsewhere in this code, and may not always be caught.  The
        caller should be prepared to deal with it.
        """
        try:
            return eval(expression, self.workspace)
        except (SystemExit, KeyboardInterrupt):            
            raise
        except:           
            self.showtraceback()              
            
    def showsyntaxerror(self, filename=None):
        """
        Display the syntax error that just occurred.

        This doesn't display a stack trace because there isn't one.

        If a filename is given, it is stuffed in the exception instead
        of what was there before (because Python's parser always uses
        "<string>" when reading from a string).

        The output is written by self.write(), below.
        """
        type, value, tb = sys.exc_info()
        sys.last_type = type
        sys.last_value = value
        sys.last_traceback = tb
        if filename and type is SyntaxError:
            # Work hard to stuff the correct filename in the exception
            try:
                msg, (dummy_filename, lineno, offset, line) = value.args
            except ValueError:
                # Not the format we expect; leave it alone
                pass
            else:
                # Stuff in the right filename
                value = SyntaxError(msg, (filename, lineno, offset, line))
                sys.last_value = value
        
        lines = traceback.format_exception_only(type, value)
        self.write_error(''.join(lines))
        
    def showtraceback(self):
        """
        Display the exception that just occurred.

        We remove the first stack item because it is our own code.

        The output is written by self.write(), below.
        """
        type, value, tb = sys.exc_info()
        sys.last_type = type
        sys.last_value = value
        sys.last_traceback = tb
        tblist = traceback.extract_tb(tb)
        del tblist[:1]
        lines = traceback.format_list(tblist)
        if lines:
            lines.insert(0, "Traceback (most recent call last):\n")
        lines.extend(traceback.format_exception_only(type, value))

        self.write_error(''.join(lines)) 

    def write_error(self, text):
        sys.stderr.write(text)
        
            
    def use_one_command(self, cmd):                        
        error_code, code = self.compile_source(cmd, symbol='auto')
        
        if error_code != 0:  
            return error_code, cmd
            
        else:
            return error_code, self.exec_code(code)
            
    def async_break(self):
        async_break(self.thread_id)  

    def async_system_exit(self):
        async_system_exit(self.thread_id) 
        
    def get_code_frames(self, back):
        frame = sys._current_frames()[self.thread_id]
        yield frame
        for _ in range(back):
            if frame.f_back is None:
                return
            frame = frame.f_back
            yield frame         

def async_raise(thread_id, exctype):
        """
        Raise the exception to the thread tid, performs cleanup if needed.
        """
        #the async exception is not immediate; it is checked every 100
        #bytecodes (=sys.getcheckinterval()).
        # if not inspect.isclass(exctype):
            # exctype = type(exctype)
        
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(thread_id), ctypes.py_object(exctype))
        
        #receiver can clear errob y ctypes.pythonapi.PyErr_Clear()?
        if res == 0:
            raise ValueError("invalid thread id")
            
        elif res != 1:
            # """if it returns a number greater than one, you're in trouble,
            # and you should call it again with exc=NULL to revert the effect"""
            ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, None)
            raise SystemError("PyThreadState_SetAsyncExc failed")
            
        else:
            pass
            #print(f'{exctype} sent to thread id {thread_id}')

def async_break(thread_id):
    async_raise(thread_id, KeyboardInterrupt)
    
def async_system_exit(thread_id):
    async_raise(thread_id, SystemExit)    



from ... import config

if config.get('qapp', False):
    from .consolepanel import MainThreadConsole, SubThreadConsole, ChildProcessConsole, ChildThreadConsole

from .consoleproxy import ConsoleGuiProxy


import re
import fnmatch
import pathlib

from qtpy import QtCore, QtGui, QtWidgets

from .base import getMap

class PathList(QtWidgets.QListWidget):

    def __init__(self, parent, paths):
        super().__init__(parent=parent)
        
        for item in paths:
            list_item = QtWidgets.QListWidgetItem(str(item))
            list_item.setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)
            self.addItem(list_item)           

class EditPaths(QtWidgets.QDialog):

    def __init__(self, paths):
        super().__init__(None)
        self.initgui(paths)
        
    def initgui(self, paths):
        self.paths = paths
        self.selectionlist = PathList(self, paths)
        
        self.addPathBtn = QtWidgets.QPushButton('Add', self)
        self.addPathBtn.clicked.connect(self.add)
        self.editPathBtn = QtWidgets.QPushButton('Edit', self)
        self.editPathBtn.clicked.connect(self.edit)
        self.delPathBtn = QtWidgets.QPushButton('Delete', self)
        self.delPathBtn.clicked.connect(self.delete)
        
        self.okBtn = QtWidgets.QPushButton('Ok', self)
        self.okBtn.clicked.connect(self.ok)
        self.cancelBtn = QtWidgets.QPushButton('Cancel', self)                       
        self.cancelBtn.clicked.connect(self.cancel)

        layout = QtWidgets.QVBoxLayout()
        
        hlayout = QtWidgets.QHBoxLayout()        
        hlayout.addWidget(self.addPathBtn)
        hlayout.addWidget(self.editPathBtn)
        hlayout.addWidget(self.delPathBtn)
        layout.addLayout(hlayout)            
        
        hlayout = QtWidgets.QHBoxLayout()
        layout.addLayout(hlayout)        
        layout.addWidget(self.selectionlist)
        
        hlayout = QtWidgets.QHBoxLayout()        
        hlayout.addWidget(self.okBtn)
        hlayout.addWidget(self.cancelBtn)
        layout.addLayout(hlayout)    
            
        self.setLayout(layout)         
        
    def add(self):
        path = pathlib.Path(getMap())
        item = QtWidgets.QListWidgetItem(str(path))
        item.setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)
        self.selectionlist.addItem(item)     
        
    def edit(self):
        for item in self.selectionlist.selectedItems():            
            path = item.text()
            path = pathlib.Path(getMap(path))
            item.setText(path)

    def delete(self):
        for item in self.selectionlist.selectedItems():        
            self.selectionlist.takeItem(self.selectionlist.row(item))
            
    def ok(self):
        self.updatePaths()
        self.done(QtWidgets.QDialog.DialogCode.Accepted)
        
    def cancel(self):
        self.done(QtWidgets.QDialog.DialogCode.Rejected)
        
    def updatePaths(self):
        self.paths.clear()
        for index in range(self.selectionlist.count()):        
            item = self.selectionlist.item(index)
            self.paths.append(item.text())

